import time
import math
import cupy as cp
import numpy as np
from sfft.utils.ReadWCS import Read_WCS

__last_update__ = "2024-09-22"
__author__ = "Lei Hu <leihu@andrew.cmu.edu> & Shu Liu <shl159@pitt.edu>"

class Cupy_WCS_Transform:
    def __init__(self):
        __author__ = "Shu Liu <shl159@pitt.edu> & Lei Hu <leihu@andrew.cmu.edu>"
        return None

    def read_cd_wcs(self, hdr_wcs, CDKEY="CD"):
        """ Read WCS information from FITS header """
        assert hdr_wcs["CTYPE1"] == "RA---TAN"
        assert hdr_wcs["CTYPE2"] == "DEC--TAN"

        N0 = int(hdr_wcs["NAXIS1"])
        N1 = int(hdr_wcs["NAXIS2"])

        CRPIX1 = float(hdr_wcs["CRPIX1"])
        CRPIX2 = float(hdr_wcs["CRPIX2"])

        CRVAL1 = float(hdr_wcs["CRVAL1"])
        CRVAL2 = float(hdr_wcs["CRVAL2"])

        LONPOLE = float(hdr_wcs["LONPOLE"])

        KEYDICT = {
            "N0": N0, "N1": N1, 
            "CRPIX1": CRPIX1, "CRPIX2": CRPIX2,
            "CRVAL1": CRVAL1, "CRVAL2": CRVAL2,
            "LONPOLE": LONPOLE
        }

        CD1_1 = hdr_wcs[f"{CDKEY}1_1"]
        CD1_2 = hdr_wcs[f"{CDKEY}1_2"] if f"{CDKEY}1_2" in hdr_wcs else 0.
        CD2_1 = hdr_wcs[f"{CDKEY}2_1"] if f"{CDKEY}2_1" in hdr_wcs else 0.
        CD2_2 = hdr_wcs[f"{CDKEY}2_2"]

        CD_GPU = cp.array([
            [CD1_1, CD1_2], 
            [CD2_1, CD2_2]
        ], dtype=cp.float64)
        
        return KEYDICT, CD_GPU

    def read_sip_wcs(self, hdr_wcs, CDKEY="CD"):
        """ Read TAN-SIP WCS information from FITS header """
        assert hdr_wcs["CTYPE1"] == "RA---TAN-SIP"
        assert hdr_wcs["CTYPE2"] == "DEC--TAN-SIP"

        N0 = int(hdr_wcs["NAXIS1"])
        N1 = int(hdr_wcs["NAXIS2"])

        CRPIX1 = float(hdr_wcs["CRPIX1"])
        CRPIX2 = float(hdr_wcs["CRPIX2"])

        CRVAL1 = float(hdr_wcs["CRVAL1"])
        CRVAL2 = float(hdr_wcs["CRVAL2"])
        
        LONPOLE = float(hdr_wcs["LONPOLE"])

        A_ORDER = int(hdr_wcs["A_ORDER"])
        B_ORDER = int(hdr_wcs["B_ORDER"])

        KEYDICT = {
            "N0": N0, "N1": N1, 
            "CRPIX1": CRPIX1, "CRPIX2": CRPIX2,
            "CRVAL1": CRVAL1, "CRVAL2": CRVAL2, 
            "LONPOLE": LONPOLE,
            "A_ORDER": A_ORDER, "B_ORDER": B_ORDER
        }

        CD1_1 = hdr_wcs[f"{CDKEY}1_1"]
        CD1_2 = hdr_wcs[f"{CDKEY}1_2"] if f"{CDKEY}1_2" in hdr_wcs else 0.
        CD2_1 = hdr_wcs[f"{CDKEY}2_1"] if f"{CDKEY}2_1" in hdr_wcs else 0.
        CD2_2 = hdr_wcs[f"{CDKEY}2_2"]

        CD_GPU = cp.array([
            [CD1_1, CD1_2], 
            [CD2_1, CD2_2]
        ], dtype=cp.float64)

        A_SIP_GPU = cp.zeros((A_ORDER+1, A_ORDER+1), dtype=cp.float64)
        B_SIP_GPU = cp.zeros((B_ORDER+1, B_ORDER+1), dtype=cp.float64)

        for p in range(A_ORDER + 1):
            for q in range(0, A_ORDER - p + 1):
                keyword = f"A_{p}_{q}"
                if keyword in hdr_wcs:
                    A_SIP_GPU[p, q] = hdr_wcs[keyword]
        
        for p in range(B_ORDER + 1):
            for q in range(0, B_ORDER - p + 1):
                keyword = f"B_{p}_{q}"
                if keyword in hdr_wcs:
                    B_SIP_GPU[p, q] = hdr_wcs[keyword]
        
        return KEYDICT, CD_GPU, A_SIP_GPU, B_SIP_GPU

    def cd_transform(self, IMAGE_X_GPU, IMAGE_Y_GPU, CD_GPU):
        """CD matrix transformation from image to world"""
        WORLD_X_GPU, WORLD_Y_GPU = cp.matmul(CD_GPU, cp.array([IMAGE_X_GPU, IMAGE_Y_GPU]))
        return WORLD_X_GPU, WORLD_Y_GPU
    
    def cd_transform_inv(self, WORLD_X_GPU, WORLD_Y_GPU, CD_GPU):
        """CD matrix transformation from world to image"""
        # CD_inv_GPU = cp.linalg.inv(CD_GPU)
        CD_inv_GPU = 1. / (CD_GPU[0, 0] * CD_GPU[1, 1] - CD_GPU[0, 1] * CD_GPU[1, 0]) * \
            cp.array([[CD_GPU[1, 1], -CD_GPU[0, 1]], [-CD_GPU[1, 0], CD_GPU[0, 0]]])
        IMAGE_X_GPU, IMAGE_Y_GPU = cp.matmul(CD_inv_GPU, cp.array([WORLD_X_GPU, WORLD_Y_GPU]))
        return IMAGE_X_GPU, IMAGE_Y_GPU
    
    def sip_forward_transform(self, u_GPU, v_GPU, A_SIP_GPU, B_SIP_GPU):
        """Forward SIP transformation from (u, v) to (U, V)"""
        A_ORDER = A_SIP_GPU.shape[0] - 1
        U_GPU = u_GPU + cp.zeros_like(u_GPU)
        for p in range(A_ORDER + 1):
            for q in range(A_ORDER - p + 1):
                U_GPU += A_SIP_GPU[p, q] * u_GPU**p * v_GPU**q
            
        B_ORDER = B_SIP_GPU.shape[0] - 1
        V_GPU = v_GPU + cp.zeros_like(v_GPU)
        for p in range(B_ORDER + 1):
            for q in range(B_ORDER - p + 1):
                V_GPU += B_SIP_GPU[p, q] * u_GPU**p * v_GPU**q
        return U_GPU, V_GPU

    def sip_backward_matrix(self, U_GPU, V_GPU, ORDER):
        """Compute the backward matrix P_UV (FP_UV or GP_UV)"""
        NSAMP = len(U_GPU)
        P_UV_GPU = cp.zeros((NSAMP, (ORDER+1)*(ORDER+2)//2), dtype=cp.float64)
        idx = 0
        for p in range(ORDER + 1):
            for q in range(ORDER - p + 1):
                P_UV_GPU[:, idx] = U_GPU**p * V_GPU**q
                idx += 1
        return P_UV_GPU

    def lstsq_sip_backward_transform(self, u_GPU, v_GPU, U_GPU, V_GPU, A_ORDER, B_ORDER):
        """Compute the backward transformation from (U, V) to (u, v)"""
        FP_UV_GPU = self.sip_backward_matrix(U_GPU=U_GPU, V_GPU=V_GPU, ORDER=A_ORDER)
        GP_UV_GPU = self.sip_backward_matrix(U_GPU=U_GPU, V_GPU=V_GPU, ORDER=B_ORDER)

        AP_lstsq_GPU = cp.linalg.lstsq(FP_UV_GPU, (u_GPU - U_GPU).reshape(-1, 1), rcond=None)[0]
        BP_lstsq_GPU = cp.linalg.lstsq(GP_UV_GPU, (v_GPU - V_GPU).reshape(-1, 1), rcond=None)[0]
        return FP_UV_GPU, GP_UV_GPU, AP_lstsq_GPU, BP_lstsq_GPU

    def sip_backward_transform(self, U_GPU, V_GPU, FP_UV_GPU, GP_UV_GPU, AP_GPU, BP_GPU):
        """Backward transformation from (U, V) to (u, v)"""
        u_GPU = U_GPU + cp.matmul(FP_UV_GPU, AP_GPU)[:, 0]
        v_GPU = V_GPU + cp.matmul(GP_UV_GPU, BP_GPU)[:, 0]
        return u_GPU, v_GPU

    def random_coord_sampling(self, N0, N1, CRPIX1, CRPIX2, NSAMP=1024, RANDOM_SEED=10086):
        """Random sampling of coordinates"""
        if RANDOM_SEED is not None:
            cp.random.seed(RANDOM_SEED)
        u_GPU = cp.random.uniform(0.5, N0+0.5, NSAMP, dtype=cp.float64) - CRPIX1
        v_GPU = cp.random.uniform(0.5, N1+0.5, NSAMP, dtype=cp.float64) - CRPIX2
        return u_GPU, v_GPU

class Cupy_Resampling:
    def __init__(self, RESAMP_METHOD="BILINEAR", VERBOSE_LEVEL=2):
        __author__ = "Lei Hu <leihu@andrew.cmu.edu>"
        self.RESAMP_METHOD = RESAMP_METHOD
        self.VERBOSE_LEVEL = VERBOSE_LEVEL
        return None
    
    def spherical_transformation(self, x_GPU, y_GPU, alpha_p, delta_p, phi_p):
        """Spherical transformation (Spherical projection & Spherical coordinate rotation) using Cupy"""
        # NOTE: we assumpe all variables are in radians!
        
        # * Spherical projection (x, y) > (phi, theta)
        phi_GPU = cp.arctan2(x_GPU, -y_GPU)
        theta_GPU = cp.arctan(1./cp.sqrt(x_GPU**2 + y_GPU**2))

        # * Spherical coordinate rotation (phi, theta) > (alpha, delta)
        delta_GPU = cp.arcsin(
            cp.sin(theta_GPU) * cp.sin(delta_p) + \
            cp.cos(theta_GPU) * cp.cos(delta_p) * cp.cos((phi_GPU - phi_p))
        )
        
        COS_TERM_GPU = (cp.sin(theta_GPU) * cp.cos(delta_p) - \
            cp.cos(theta_GPU) * cp.sin(delta_p) * cp.cos(phi_GPU - phi_p)) / cp.cos(delta_GPU)
        SIN_TERM_GPU = - cp.cos(theta_GPU) * cp.sin(phi_GPU - phi_p) / cp.cos(delta_GPU)
        alpha_GPU = alpha_p + cp.arctan2(SIN_TERM_GPU, COS_TERM_GPU)

        return alpha_GPU, delta_GPU

    def inverse_spherical_transformation(self, alpha_GPU, delta_GPU, alpha_p, delta_p, phi_p):
        """Inverse Spherical transformation (Spherical projection & Spherical coordinate rotation) using Cupy"""
        # NOTE: we assumpe all variables are in radians!
        
        # * Inverse Spherical coordinate rotation (alpha, delta) > (phi, theta)
        theta_GPU = cp.arcsin(cp.sin(delta_GPU) * cp.sin(delta_p) + \
            cp.cos(delta_GPU) * cp.cos(delta_p) * cp.cos(alpha_GPU - alpha_p))

        COS_TERM_GPU = (cp.sin(delta_GPU) * cp.cos(delta_p) - \
            cp.cos(delta_GPU) * cp.sin(delta_p) * cp.cos(alpha_GPU - alpha_p)) / cp.cos(theta_GPU)
        SIN_TERM_GPU = - cp.cos(delta_GPU) * cp.sin(alpha_GPU - alpha_p) / cp.cos(theta_GPU)
        phi_GPU = phi_p + cp.arctan2(SIN_TERM_GPU, COS_TERM_GPU)

        # * Inverse Spherical projection (phi, theta) > (x, y)
        r_GPU = 1./cp.tan(theta_GPU)
        x_GPU = r_GPU * cp.sin(phi_GPU)
        y_GPU = - r_GPU * cp.cos(phi_GPU)

        return x_GPU, y_GPU

    def resamp_projection_sip(self, hdr_obj, hdr_targ, CDKEY="CD", NSAMP=1024, RANDOM_SEED=10086):
        """Project the target pixel centers to the object frame using Cupy for TAN-SIP WCS"""
        NTX = int(hdr_targ["NAXIS1"]) 
        NTY = int(hdr_targ["NAXIS2"])

        XX_targ_GPU, YY_targ_GPU = cp.meshgrid(
            cp.arange(0, NTX) + 1., 
            cp.arange(0, NTY) + 1., 
            indexing='ij'
        )
        
        CWT = Cupy_WCS_Transform()
        if True:
            # read header | target
            KEYDICT, CD_GPU, A_SIP_GPU, B_SIP_GPU = CWT.read_sip_wcs(hdr_wcs=hdr_targ, CDKEY=CDKEY)
            CRPIX1_targ, CRPIX2_targ = KEYDICT["CRPIX1"], KEYDICT["CRPIX2"]
            CRVAL1_targ, CRVAL2_targ = KEYDICT["CRVAL1"], KEYDICT["CRVAL2"]
            LONPOLE_targ = KEYDICT["LONPOLE"]
            
            # forward transformation 0: (p1, p2) > (u, v) | target
            u_GPU, v_GPU = XX_targ_GPU.flatten() - CRPIX1_targ, YY_targ_GPU.flatten() - CRPIX2_targ
            
            # forward transformation 1: (u, v) > (x, y) | target
            U_GPU, V_GPU = CWT.sip_forward_transform(u_GPU=u_GPU, v_GPU=v_GPU, A_SIP_GPU=A_SIP_GPU, B_SIP_GPU=B_SIP_GPU)
            x_GPU, y_GPU = CWT.cd_transform(IMAGE_X_GPU=U_GPU, IMAGE_Y_GPU=V_GPU, CD_GPU=CD_GPU)

            # forward transformation 2 & 3 to sky: (x, y) > (phi, theta) > (alpha, delta) | target
            ra_GPU, dec_GPU = self.spherical_transformation(
                x_GPU=cp.deg2rad(x_GPU), 
                y_GPU=cp.deg2rad(y_GPU), 
                alpha_p=cp.deg2rad(CRVAL1_targ), 
                delta_p=cp.deg2rad(CRVAL2_targ),
                phi_p=cp.deg2rad(LONPOLE_targ)
            )   # in radians!
        
        if True:
            # read header | object
            KEYDICT, CD_GPU, A_SIP_GPU, B_SIP_GPU = CWT.read_sip_wcs(hdr_wcs=hdr_obj, CDKEY=CDKEY)
            CRPIX1_obj, CRPIX2_obj = KEYDICT["CRPIX1"], KEYDICT["CRPIX2"]
            CRVAL1_obj, CRVAL2_obj = KEYDICT["CRVAL1"], KEYDICT["CRVAL2"]
            LONPOLE_obj = KEYDICT["LONPOLE"]

            # inverse transformation 3 & 2: (alpha, delta) > (phi, theta) > (x, y) | object
            x_GPU, y_GPU = self.inverse_spherical_transformation(
                alpha_GPU=ra_GPU, 
                delta_GPU=dec_GPU, 
                alpha_p=cp.deg2rad(CRVAL1_obj), 
                delta_p=cp.deg2rad(CRVAL2_obj),
                phi_p=cp.deg2rad(LONPOLE_obj)
            )   # in radians!

            # inverse transformation 1: (x, y) > (u, v) | object
            # inverse CD transformation
            U_GPU, V_GPU = CWT.cd_transform_inv(
                WORLD_X_GPU=cp.rad2deg(x_GPU), 
                WORLD_Y_GPU=cp.rad2deg(y_GPU), 
                CD_GPU=CD_GPU
            )

            # inverse SIP transformation
            def FIT_INVERSE_SIP(KEYDICT, A_SIP_GPU, B_SIP_GPU, NSAMP, RANDOM_SEED):
                N0, N1 = KEYDICT["N0"], KEYDICT["N1"]
                CRPIX1, CRPIX2 = KEYDICT["CRPIX1"], KEYDICT["CRPIX2"]
                A_ORDER, B_ORDER = KEYDICT["A_ORDER"], KEYDICT["B_ORDER"]
                
                # sampling random coordinates
                u_GPU, v_GPU = CWT.random_coord_sampling(N0=N0, N1=N1, 
                    CRPIX1=CRPIX1, CRPIX2=CRPIX2, NSAMP=NSAMP, RANDOM_SEED=RANDOM_SEED)
                
                # fit polynomial form backward transformation | object
                U_GPU, V_GPU = CWT.sip_forward_transform(u_GPU=u_GPU, v_GPU=v_GPU, A_SIP_GPU=A_SIP_GPU, B_SIP_GPU=B_SIP_GPU)
                AP_lstsq_GPU, BP_lstsq_GPU = CWT.lstsq_sip_backward_transform(u_GPU=u_GPU, v_GPU=v_GPU, 
                    U_GPU=U_GPU, V_GPU=V_GPU, A_ORDER=A_ORDER, B_ORDER=B_ORDER)[2:4]
                return AP_lstsq_GPU, BP_lstsq_GPU

            AP_lstsq_GPU, BP_lstsq_GPU = FIT_INVERSE_SIP(
                KEYDICT=KEYDICT, A_SIP_GPU=A_SIP_GPU, B_SIP_GPU=B_SIP_GPU,
                NSAMP=NSAMP, RANDOM_SEED=RANDOM_SEED
            )

            FP_UV_GPU = CWT.sip_backward_matrix(U_GPU=U_GPU, V_GPU=V_GPU, ORDER=KEYDICT["A_ORDER"])
            GP_UV_GPU = CWT.sip_backward_matrix(U_GPU=U_GPU, V_GPU=V_GPU, ORDER=KEYDICT["B_ORDER"])

            u_GPU, v_GPU = CWT.sip_backward_transform(U_GPU=U_GPU, V_GPU=V_GPU, 
                FP_UV_GPU=FP_UV_GPU, GP_UV_GPU=GP_UV_GPU, AP_GPU=AP_lstsq_GPU, BP_GPU=BP_lstsq_GPU)

            # inverse transformation 0: (u, v) > (p1, p2) | object
            XX_proj_GPU = CRPIX1_obj + u_GPU.reshape((NTX, NTY))
            YY_proj_GPU = CRPIX2_obj + v_GPU.reshape((NTX, NTY))

        return XX_proj_GPU, YY_proj_GPU

    def resamp_projection_astropy(self, hdr_obj, hdr_targ):
        """Mapping the target pixel centers to the object frame using Astropy"""
        # * read object WCS and target WCS
        w_obj = Read_WCS.RW(hdr=hdr_obj, VERBOSE_LEVEL=self.VERBOSE_LEVEL)
        w_targ = Read_WCS.RW(hdr=hdr_targ, VERBOSE_LEVEL=self.VERBOSE_LEVEL)
        
        # * maaping target pixel centers to the object frame
        NTX = int(hdr_targ["NAXIS1"]) 
        NTY = int(hdr_targ["NAXIS2"])
        
        XX_targ, YY_targ = np.meshgrid(np.arange(0, NTX)+1., np.arange(0, NTY)+1., indexing='ij')
        XY_targ = np.array([XX_targ.flatten(), YY_targ.flatten()]).T
        XY_proj = w_obj.all_world2pix(w_targ.all_pix2world(XY_targ, 1), 1)

        XX_proj = XY_proj[:, 0].reshape((NTX, NTY))
        YY_proj = XY_proj[:, 1].reshape((NTX, NTY))

        XX_proj_GPU = cp.array(XX_proj, dtype=np.float64)
        YY_proj_GPU = cp.array(YY_proj, dtype=np.float64)

        return XX_proj_GPU, YY_proj_GPU

    def frame_extension(self, XX_proj_GPU, YY_proj_GPU, PixA_obj_GPU, PAD_FILL_VALUE=0., NAN_FILL_VALUE=0.):
        """Extend the object frame for resampling"""
        NTX, NTY = XX_proj_GPU.shape
        NOX, NOY = PixA_obj_GPU.shape

        # * padding the object frame
        if self.RESAMP_METHOD == 'BILINEAR':
            KERHW = (1, 1)

        if self.RESAMP_METHOD == 'LANCZOS3':
            KERHW = (3, 3)

        # find the root index and shift with maximal kernel halfwidth (KERHW)
        # Note: the index ranges (RMIN --- RMAX) and (CMIN --- CMAX) can cover
        #       all pixels that interpolation may be use.

        RMIN = (math.floor(XX_proj_GPU.min().item()) - 1) - KERHW[0]
        RMAX = (math.floor(XX_proj_GPU.max().item()) - 1) + KERHW[0]
        RPAD = (-np.min([RMIN, 0]), np.max([RMAX - (NOX - 1), 0]))

        CMIN = (math.floor(YY_proj_GPU.min().item()) - 1) - KERHW[1]
        CMAX = (math.floor(YY_proj_GPU.max().item()) - 1) + KERHW[1]
        CPAD = (-np.min([CMIN, 0]), np.max([CMAX - (NOY - 1), 0]))

        PAD_WIDTH = (RPAD, CPAD)    
        PixA_Eobj_GPU = cp.pad(PixA_obj_GPU, PAD_WIDTH, mode='constant', constant_values=PAD_FILL_VALUE)
        if NAN_FILL_VALUE is not None:
            PixA_Eobj_GPU[cp.isnan(PixA_Eobj_GPU)] = NAN_FILL_VALUE
        NEOX, NEOY = PixA_Eobj_GPU.shape
        
        XX_Eproj_GPU = XX_proj_GPU + PAD_WIDTH[0][0]
        YY_Eproj_GPU = YY_proj_GPU + PAD_WIDTH[1][0]

        RMIN_E = (math.floor(XX_Eproj_GPU.min().item()) - 1) - KERHW[0]
        RMAX_E = (math.floor(XX_Eproj_GPU.max().item()) - 1) + KERHW[0]

        CMIN_E = (math.floor(YY_Eproj_GPU.min().item()) - 1) - KERHW[1]
        CMAX_E = (math.floor(YY_Eproj_GPU.max().item()) - 1) + KERHW[1]

        assert RMIN_E >= 0 and CMIN_E >= 0 
        assert RMAX_E < NEOX and CMAX_E < NEOY

        EProjDict = {}
        EProjDict["NTX"] = NTX
        EProjDict["NTY"] = NTY

        EProjDict["NOX"] = NOX
        EProjDict["NOY"] = NOY

        EProjDict["NEOX"] = NEOX
        EProjDict["NEOY"] = NEOY
        EProjDict["KERHW"] = KERHW

        EProjDict["XX_Eproj_GPU"] = XX_Eproj_GPU
        EProjDict["YY_Eproj_GPU"] = YY_Eproj_GPU

        return PixA_Eobj_GPU, EProjDict
    
    def resampling(self, PixA_Eobj_GPU, EProjDict, PIXEL_SCALE_FACTOR=1., THREAD_PER_BLOCK=8, USE_SHARED_MEMORY=False):
        """Resampling the object frame to the target frame using Cupy"""
        NTX = EProjDict["NTX"]
        NTY = EProjDict["NTY"]

        NEOX = EProjDict["NEOX"]
        NEOY = EProjDict["NEOY"]
        KERHW = EProjDict["KERHW"]

        XX_Eproj_GPU = EProjDict["XX_Eproj_GPU"]
        YY_Eproj_GPU = EProjDict["YY_Eproj_GPU"]

        # * Cupy configuration
        GPUManage = lambda NT: ((NT-1)//THREAD_PER_BLOCK + 1, min(NT, THREAD_PER_BLOCK))
        BpG_PIX0, TpB_PIX0 = GPUManage(NTX)
        BpG_PIX1, TpB_PIX1 = GPUManage(NTY)
        BpG_PIX, TpB_PIX = (BpG_PIX0, BpG_PIX1), (TpB_PIX0, TpB_PIX1, 1)

        PixA_resamp_GPU = cp.zeros((NTX, NTY), dtype=np.float64)

        if self.RESAMP_METHOD == "BILINEAR":

            if USE_SHARED_MEMORY:
                print("MeLOn WARNING: Shared memory is not implemented for bilinear resampling!")

            # * perform bilinear resampling using CUDA
            # input: PixA_Eobj | (NEOX, NEOY)
            # input: XX_Eproj, YY_Eproj | (NTX, NTY)
            # output: PixA_resamp | (NTX, NTY)
            
            _refdict = {'NTX': NTX, 'NTY': NTY, 'NEOX': NEOX, 'NEOY': NEOY}
            _funcstr = r"""
            extern "C" __global__ void kmain(double XX_Eproj_GPU[%(NTX)s][%(NTY)s], double YY_Eproj_GPU[%(NTX)s][%(NTY)s], 
                double PixA_Eobj_GPU[%(NEOX)s][%(NEOY)s], double PixA_resamp_GPU[%(NTX)s][%(NTY)s])
            {
                int ROW = blockIdx.x*blockDim.x+threadIdx.x;
                int COL = blockIdx.y*blockDim.y+threadIdx.y;

                int NTX = %(NTX)s;
                int NTY = %(NTY)s;
                int NEOX = %(NEOX)s;
                int NEOY = %(NEOY)s;

                if (ROW < NTX && COL < NTY) {
                
                    double x = XX_Eproj_GPU[ROW][COL];
                    double y = YY_Eproj_GPU[ROW][COL];

                    int r1 = floor(x) - 1;
                    int c1 = floor(y) - 1;
                    int r2 = r1 + 1;
                    int c2 = c1 + 1;

                    double dx = x - floor(x); 
                    double dy = y - floor(y);

                    double w11 = (1-dx) * (1-dy);
                    double w12 = (1-dx) * dy;
                    double w21 = dx * (1-dy);
                    double w22 = dx * dy;

                    PixA_resamp_GPU[ROW][COL] = w11 * PixA_Eobj_GPU[r1][c1] + w12 * PixA_Eobj_GPU[r1][c2] + 
                        w21 * PixA_Eobj_GPU[r2][c1] + w22 * PixA_Eobj_GPU[r2][c2];
                }
            }
            """
            _code = _funcstr % _refdict
            _module = cp.RawModule(code=_code, backend=u'nvcc', translate_cucomplex=False)
            resamp_func = _module.get_function('kmain')
            
            t0 = time.time()
            resamp_func(args=(XX_Eproj_GPU, YY_Eproj_GPU, PixA_Eobj_GPU, PixA_resamp_GPU), block=TpB_PIX, grid=BpG_PIX)
            PixA_resamp_GPU *= PIXEL_SCALE_FACTOR
            
            if self.VERBOSE_LEVEL in [1, 2]:
                print('MeLOn CheckPoint: Cuda resampling takes [%.6f s]' %(time.time() - t0))

        return PixA_resamp_GPU

class Cupy_ZoomRotate:
    @staticmethod
    def CZR(PixA_obj_GPU, ZOOM_SCALE_X=1., ZOOM_SCALE_Y=1., OUTSIZE_PARIRY_X='UNCHANGED', OUTSIZE_PARIRY_Y='UNCHANGED', \
        PATTERN_ROTATE_ANGLE=0.0, RESAMP_METHOD='BILINEAR', PAD_FILL_VALUE=0., NAN_FILL_VALUE=0., \
        THREAD_PER_BLOCK=8, USE_SHARED_MEMORY=False, VERBOSE_LEVEL=2):
        """ Zoom and rotate an image using Cupy """
        __author__ = "Lei Hu <leihu@andrew.cmu.edu>"
        
        # check inputs
        assert ZOOM_SCALE_X > 0.0
        assert ZOOM_SCALE_Y > 0.0
        assert 0.0 <= PATTERN_ROTATE_ANGLE < 360.0
        NAXIS1_ORI, NAXIS2_ORI = PixA_obj_GPU.shape

        # * determine the number of pixels of the zoomed image
        def NPIX_ORI2ZOOMED(NPIX_ORI, ZOOM_SCALE, OUTSIZE_PARIRY):
            if OUTSIZE_PARIRY == 'UNCHANGED':
                if NPIX_ORI % 2 == 0: OUTSIZE_UPARIRY = 'EVEN'
                if NPIX_ORI % 2 == 1: OUTSIZE_UPARIRY = 'ODD'
            if OUTSIZE_PARIRY == 'ODD': OUTSIZE_UPARIRY = 'ODD'
            if OUTSIZE_PARIRY == 'EVEN': OUTSIZE_UPARIRY = 'EVEN'
            HALFWIDTH_ORI = NPIX_ORI/2.0  # unit of original pixel
            if OUTSIZE_UPARIRY == 'EVEN':
                NPIX_ZOOMED = 2 * math.ceil(HALFWIDTH_ORI/ZOOM_SCALE)
            if OUTSIZE_UPARIRY == 'ODD':
                NPIX_ZOOMED = 2 * math.ceil((HALFWIDTH_ORI - ZOOM_SCALE/2.0)/ZOOM_SCALE) + 1
            return NPIX_ZOOMED

        NAXIS1_ZOOMED = NPIX_ORI2ZOOMED(NPIX_ORI=NAXIS1_ORI, ZOOM_SCALE=ZOOM_SCALE_X, OUTSIZE_PARIRY=OUTSIZE_PARIRY_X)
        NAXIS2_ZOOMED = NPIX_ORI2ZOOMED(NPIX_ORI=NAXIS2_ORI, ZOOM_SCALE=ZOOM_SCALE_Y, OUTSIZE_PARIRY=OUTSIZE_PARIRY_Y)

        def BACKWARD_TRANSFORM(X_ROTATED_GPU, Y_ROTATED_GPU):
            # * prepare for the frame transform
            CRPIX1_ZOOMED = 0.5 + NAXIS1_ZOOMED/2.
            CRPIX2_ZOOMED = 0.5 + NAXIS2_ZOOMED/2.

            ALPHA_INV = cp.deg2rad(-PATTERN_ROTATE_ANGLE)
            ROTMAT_INV_GPU = cp.array([
                [cp.cos(ALPHA_INV), -cp.sin(ALPHA_INV)], 
                [cp.sin(ALPHA_INV), cp.cos(ALPHA_INV)]
            ])

            # * frame transform from rotated to zoomed
            COORD_ROTATED_GPU = cp.array([
                X_ROTATED_GPU - CRPIX1_ZOOMED,
                Y_ROTATED_GPU - CRPIX2_ZOOMED
            ])
            
            COORD_ZOOMED_GPU = cp.dot(ROTMAT_INV_GPU, COORD_ROTATED_GPU)
            COORD_ZOOMED_GPU[0, :] += CRPIX1_ZOOMED
            COORD_ZOOMED_GPU[1, :] += CRPIX2_ZOOMED
            X_ZOOMED_GPU, Y_ZOOMED_GPU = COORD_ZOOMED_GPU
            
            # * frame transform from zoomed to original
            OFFSET_X = ZOOM_SCALE_X * (0.5 + NAXIS1_ZOOMED / 2.0) - (0.5 + NAXIS1_ORI / 2.0)
            OFFSET_Y = ZOOM_SCALE_Y * (0.5 + NAXIS2_ZOOMED / 2.0) - (0.5 + NAXIS2_ORI / 2.0)

            X_ORI_GPU = X_ZOOMED_GPU * ZOOM_SCALE_X - OFFSET_X
            Y_ORI_GPU = Y_ZOOMED_GPU * ZOOM_SCALE_Y - OFFSET_Y

            return X_ORI_GPU, Y_ORI_GPU

        XX_targ_GPU, YY_targ_GPU = cp.meshgrid(
            cp.arange(0, NAXIS1_ZOOMED) + 1., 
            cp.arange(0, NAXIS2_ZOOMED) + 1., 
            indexing='ij'
        )

        X_ORI_GPU, Y_ORI_GPU = BACKWARD_TRANSFORM(X_ROTATED_GPU=XX_targ_GPU.flatten(), Y_ROTATED_GPU=YY_targ_GPU.flatten())
        XX_proj_GPU = X_ORI_GPU.reshape((NAXIS1_ZOOMED, NAXIS2_ZOOMED))
        YY_proj_GPU = Y_ORI_GPU.reshape((NAXIS1_ZOOMED, NAXIS2_ZOOMED))

        GWR = Cupy_Resampling(RESAMP_METHOD=RESAMP_METHOD, VERBOSE_LEVEL=VERBOSE_LEVEL)
        PixA_Eobj_GPU, EProjDict = GWR.frame_extension(
            XX_proj_GPU=XX_proj_GPU, YY_proj_GPU=YY_proj_GPU, PixA_obj_GPU=PixA_obj_GPU,
            PAD_FILL_VALUE=PAD_FILL_VALUE, NAN_FILL_VALUE=NAN_FILL_VALUE
        )
        
        PIXEL_SCALE_FACTOR = ZOOM_SCALE_X * ZOOM_SCALE_Y
        PixA_resamp_GPU = GWR.resampling(
            PixA_Eobj_GPU=PixA_Eobj_GPU, EProjDict=EProjDict, PIXEL_SCALE_FACTOR=PIXEL_SCALE_FACTOR, 
            THREAD_PER_BLOCK=THREAD_PER_BLOCK, USE_SHARED_MEMORY=USE_SHARED_MEMORY
        )

        return PixA_resamp_GPU



import numpy as np
from copy import deepcopy
from astropy.io import fits
# version: Feb 24, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Read_SFFTSolution:

    def __init__(self):

        """
        * Notes on the representation of a spatially-varying kernel (SVK)
        # [1] Using SFFT Dictionary
        #     > SVK_xy = sum_ab (Ac_xyab * K_ab), where Ac_xyab = sum_ij (ac_ijab * x^i * y^j)
        #     > Sfft_dict[(i, j)][a, b] = ac_ijab
        #  
        #     NOTE: K_ab is modified delta basis (see the sfft paper)
        #           Ac (ac) is the symbol A (a) with circle hat in sfft paper, note that ac = a/(N0*N1).
        #           Although (x, y) are the integer indices of a pixel in sfft paper, we actually used a 
        #           scaled image coordinate (i.e., ScaledFortranCoor) of the pixel center in our implementation.
        #           Here, above equations allows for all possible image coordinates (not just pixel center).
        #
        # [2] Using 'Standard' Dictionary
        #     > SVK_xy = sum_ab (B_xyab * D_ab), where B_xyab = sum_ij (b_ijab * x^i * y^j)
        #     > Standard_dict[(i, j)][a, b] = b_ijab
        #     
        #     NOTE D_ab is the standard Cartesian-delta basis
        #          Likewise, (x, y) is ScaledFortranCoor of all possible image coordinates.
        #
        # P.S. Coordinate Example: Given Pixel Location at (row=3, column=5) in an image of size (1024, 2048)
        #      FortranCoor = (4.0, 6.0) | ScaledFortranCoor = (4.0/1024, 6.0/2048)
        #
        # P.S. In principle, a and b are in the range of [-w0, w0 + 1) and [-w1 , w1+ 1), respectively.
        #      However, Sfft_dict[(i, j)] or Standard_dict[(i, j)] has to be a matrix with 
        #      non-negative indices starting from 0. In practice, the indices were tweaked 
        #      as [a, b] > [a + w0, b + w1].
        #
        """

        pass

    def FromArray(self, Solution, N0, N1, L0, L1, DK, Fpq):

        DKx, DKy = DK, DK
        w0 = int((L0-1)/2)
        w1 = int((L1-1)/2)
        
        REF_ij = np.array([(i, j) for i in range(DKx+1) for j in range(DKy+1-i)]).astype(int)
        REF_ab = np.array([(a_pos - w0, b_pos - w1) for a_pos in range(L0) for b_pos in range(L1)]).astype(int)
        
        SREF_ijab = np.array([(ij, ab) for ij in range(REF_ij.shape[0]) for ab in range(REF_ab.shape[0])]).astype(int)
        Fijab = SREF_ijab.shape[0]

        # initialize SFFT Dictionary 
        Sfft_dict = {}
        for i in range(DKx+1):
            for j in range(DKy+1-i):
                Sfft_dict[(i, j)] = np.zeros((L0, L1)).astype(float)

        # fill the SFFT Dictionary
        a_ijab = Solution[:-Fpq]      # drop differential background
        ac_ijab = a_ijab / (N0*N1)    # a > ac

        for idx in range(Fijab):
            ij, ab = SREF_ijab[idx]
            a, b = REF_ab[ab]
            i, j = REF_ij[ij]

            # add offsets w0/w1 for non-negative indices
            Sfft_dict[(i, j)][a+w0, b+w1] = ac_ijab[idx]   
        
        return Sfft_dict

    def FromFITS(self, FITS_Solution):

        phr = fits.getheader(FITS_Solution, ext=0)
        N0, N1 = int(phr['N0']), int(phr['N1'])
        L0, L1 = int(phr['L0']), int(phr['L1'])
        DK = int(phr['DK'])
        Fpq = int(phr['FPQ'])

        Solution = fits.getdata(FITS_Solution, ext=0)[0]
        Sfft_dict = self.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)

        return Sfft_dict

class SVKDict_ST2SFFT:
    @staticmethod
    def convert(DKx, DKy, Standard_dict):
        # convert Standard-dict into Sfft-dict (Basis change)
        L0, L1 = Standard_dict[(0, 0)].shape
        w0 = int((L0-1)/2)
        w1 = int((L1-1)/2)
        Sfft_dict = deepcopy(Standard_dict)
        for i in range(DKx+1):
            for j in range(DKy+1-i):
                Sfft_dict[(i, j)][0+w0, 0+w1] = np.sum(Standard_dict[(i, j)])
        return Sfft_dict

class SVKDict_SFFT2ST:
    @staticmethod
    def convert(DKx, DKy, Sfft_dict):
        # convert Sfft-dict into Standard-dict (Basis change)
        L0, L1 = Sfft_dict[(0, 0)].shape
        w0 = int((L0-1)/2)
        w1 = int((L1-1)/2)
        Standard_dict = deepcopy(Sfft_dict)
        for i in range(DKx+1):
            for j in range(DKy+1-i):
                Standard_dict[(i, j)][0+w0, 0+w1] = 2 * Sfft_dict[(i, j)][0+w0, 0+w1] - \
                    np.sum(Sfft_dict[(i, j)])
        return Standard_dict

class Realize_MatchingKernel:

    def __init__(self, XY_q):
        self.XY_q = XY_q

    def FromArray(self, Solution, N0, N1, L0, L1, DK, Fpq):

        DKx, DKy = DK, DK
        # convert requested coordinates
        sXY_q = self.XY_q.astype(float)   # input requested coordinates in FortranCoor [global convention]
        sXY_q[:, 0] /= N0                 # convert to ScaledFortranCoor [local convention]
        sXY_q[:, 1] /= N1                 # convert to ScaledFortranCoor [local convention]

        # read SFFTSolution and convert to standard dictionary
        Sfft_dict = Read_SFFTSolution().FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)
        Standard_dict = SVKDict_SFFT2ST.convert(DKx=DKx, DKy=DKy, Sfft_dict=Sfft_dict)

        # realize kernels at given coordinates
        B = np.array([sXY_q[:, 0]**i * sXY_q[:, 1]**j for i in range(DKx+1) for j in range(DKy+1-i)])
        SSS = np.array([Standard_dict[(i, j)] for i in range(DKx+1) for j in range(DKy+1-i)])
        KerStack = np.tensordot(B, SSS, (0, 0))   # (Num_request, NX_kernel, NY_kernel)
        
        return KerStack
    
    def FromFITS(self, FITS_Solution):

        phr = fits.getheader(FITS_Solution, ext=0)
        N0, N1 = int(phr['N0']), int(phr['N1'])
        L0, L1 = int(phr['L0']), int(phr['L1'])
        DK = int(phr['DK'])
        Fpq = int(phr['FPQ'])

        Solution = fits.getdata(FITS_Solution, ext=0)[0]
        KerStack = self.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)

        return KerStack

class Realize_FluxScaling:

    def __init__(self, XY_q):
        self.XY_q = XY_q

    def FromArray(self, Solution, N0, N1, L0, L1, DK, Fpq):

        DKx, DKy = DK, DK
        w0 = int((L0-1)/2)
        w1 = int((L1-1)/2)

        # convert requested coordinates
        sXY_q = self.XY_q.astype(float)   # input requested coordinates in FortranCoor [global convention]
        sXY_q[:, 0] /= N0                 # convert to ScaledFortranCoor [local convention]
        sXY_q[:, 1] /= N1                 # convert to ScaledFortranCoor [local convention]

        # read SFFTSolution and convert to standard dictionary
        Sfft_dict = Read_SFFTSolution().FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)
        
        # calculate the flux scaling through SFFT convolution at given coordinates
        # NOTE: flux scaling Ac_xy00 = sum_ij (ac_ij00 * x^i * y^j) and ac_ijab = Sfft_dict[(i, j)][a, b].
        #       in practice, ac_ij00 = Sfft_dict[(i, j)][w0, w1]  # w0, w1 are offsets
        # NOTE: the flux scaling is consistent with the sum of matching kernel at any given coordinate.
        #       it indicates the amplification of the overall flux by the convolution in the image subtraction.

        SFFT_FSCAL_q = np.zeros(self.XY_q.shape[0]).astype(float)
        for i in range(DKx+1):
            for j in range(DKy+1-i):
                SFFT_FSCAL_q += Sfft_dict[(i, j)][w0, w1] * sXY_q[:, 0]**i * sXY_q[:, 1]**j

        return SFFT_FSCAL_q
    
    def FromFITS(self, FITS_Solution):

        phr = fits.getheader(FITS_Solution, ext=0)
        N0, N1 = int(phr['N0']), int(phr['N1'])
        L0, L1 = int(phr['L0']), int(phr['L1'])
        DK = int(phr['DK'])
        Fpq = int(phr['FPQ'])

        Solution = fits.getdata(FITS_Solution, ext=0)[0]
        SFFT_FSCAL_q = self.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)

        return SFFT_FSCAL_q



import math
import numpy as np
from sfft.utils.ConvKernelConvertion import ConvKernel_Convertion
# version: Mar 10, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class DeCorrelation_Calculator:
    @staticmethod
    def DCC(MK_JLst, SkySig_JLst, MK_ILst=[], SkySig_ILst=[], MK_Fin=None, KERatio=2.0, VERBOSE_LEVEL=2):
        
        """
        # * Remarks on Input
        #   i. Image-Stacking Mode: NumI = 0
        #      MK_Fin will not work, NumJ >= 2 and NOT all J's kernel are None
        #   ii. Image-Subtraction Mode: NumI & NumJ >= 1
        #       NOT all (I / J / Fin) kernel are None
        #
        # * Remarks on difference flip
        #   D = REF - SCI * K
        #   fD = SCI * K - REF = fREF - fSCI * K
        #   NOTE: fD and D have consistent decorrelation kernel
        #         as Var(REF) = Var(fREF) and Var(SCI) = Var(fSCI)
        #
        # * Remarks on DeCorrelation Kernel Size
        #   The DeCorrelation Kernel is derived in Fourier Space, but it is not proper to directly 
        #   perform DeCorrelation in Fourier Space (equivalently, perform a convolution with Kernel-Size = Image-Size).
        #   a. convolution (and resampling) process has very local effect, it is unnecesseary to use a large decorrelation kernel.
        #   b. if we use a large decorrelation kernel, you will find only very few engery distributed at outskirt regions, 
        #      however, the low weight can degrade the decorrelation convolution by the remote saturated pixels.
        #
        """

        NumI, NumJ = len(MK_ILst), len(MK_JLst)
        if NumI == 0: 
            Mode = 'Image-Stacking'
            if NumJ < 2: 
                _error_message = 'Image-Stacking Mode requires at least 2 J-images!'
                raise Exception('MeLOn ERROR: %s' %_error_message)
            if np.sum([MKj is not None for MKj in MK_JLst]) == 0:
                _error_message = 'Image-Stacking Mode requires at least 1 not-None J-kernel!'
                raise Exception('MeLOn ERROR: %s' %_error_message)

        if NumI >= 1:
            Mode = 'Image-Subtraction'
            if NumJ == 0: 
                _error_message = 'Image-Subtraction Mode requires at least 1 I-image & 1 J-image!'
                raise Exception('MeLOn ERROR: %s' %_error_message)
            if np.sum([MK is not None for MK in MK_JLst+MK_ILst+[MK_Fin]]) == 0:
                _error_message = 'Image-Subtraction Mode requires at least 1 not-None J/I/Fin-kernel!'
                raise Exception('MeLOn ERROR: %s' %_error_message)
        
        MK_Queue = MK_JLst.copy()
        if Mode == 'Image-Subtraction': MK_Queue += [MK_Fin] + MK_ILst
        L0_KDeCo = int(round(KERatio * np.max([MK.shape[0] for MK in MK_Queue if MK is not None])))
        L1_KDeCo = int(round(KERatio * np.max([MK.shape[1] for MK in MK_Queue if MK is not None])))
        if L0_KDeCo%2 == 0: L0_KDeCo += 1
        if L1_KDeCo%2 == 0: L1_KDeCo += 1

        if VERBOSE_LEVEL in [1, 2]:
            _message = 'DeCorrelation Kernel with size [%d, %d]' %(L0_KDeCo, L1_KDeCo)
            print('MeLOn CheckPoint: %s' %_message)

        # trivial image size, just typically larger than the kernel size.
        N0 = 2 ** (math.ceil(np.log2(np.max([MK.shape[0] for MK in MK_Queue if MK is not None])))+1)
        N1 = 2 ** (math.ceil(np.log2(np.max([MK.shape[1] for MK in MK_Queue if MK is not None])))+1)
        
        # construct the DeNonimator (a real-positive map) in Fourier Space
        uMK = np.array([[0, 0, 0], [0, 1, 0], [0, 0, 0]]).astype(float)
        def Get_JTerm(MKj, skysig):
            if MKj is not None: KIMG_CSZ = ConvKernel_Convertion.CSZ(MKj, N0, N1)
            if MKj is None: KIMG_CSZ = ConvKernel_Convertion.CSZ(uMK, N0, N1)
            kft = np.fft.fft2(KIMG_CSZ)
            kft2 = (np.conj(kft) * kft).real
            term = (skysig**2 * kft2) / NumJ**2
            return term
        
        if MK_Fin is not None: KIMG_CSZ = ConvKernel_Convertion.CSZ(MK_Fin, N0, N1)
        if MK_Fin is None: KIMG_CSZ = ConvKernel_Convertion.CSZ(uMK, N0, N1)
        kft = np.fft.fft2(KIMG_CSZ)
        kft2_Fin = (np.conj(kft) * kft).real

        def Get_ITerm(MKi, skysig):
            if MKi is not None: KIMG_CSZ = ConvKernel_Convertion.CSZ(MKi, N0, N1)
            if MKi is None: KIMG_CSZ = ConvKernel_Convertion.CSZ(uMK, N0, N1)
            kft = np.fft.fft2(KIMG_CSZ)
            kft2 = (np.conj(kft) * kft).real
            term = (skysig**2 * kft2 * kft2_Fin) / NumI**2 
            return term
        
        DeNo = 0.0
        for MKj, skysig in zip(MK_JLst, SkySig_JLst):
            DeNo += Get_JTerm(MKj, skysig)
        if Mode == 'Image-Subtraction':
            for MKi, skysig in zip(MK_ILst, SkySig_ILst):
                DeNo += Get_ITerm(MKi, skysig)

        FDeCo = np.sqrt(1.0 / DeNo)        # real & conjugate-symmetric
        DeCo = np.fft.ifft2(FDeCo).real    # no imaginary part
        KDeCo = ConvKernel_Convertion.iCSZ(DeCo, L0_KDeCo, L1_KDeCo)
        KDeCo = KDeCo / np.sum(KDeCo)      # rescale to have Unit kernel sum

        return KDeCo



import sys
import time
import math
import warnings
import threading
import numpy as np
import os.path as pa
from astropy.io import fits
from sfft.AutoCrowdedPrep import Auto_CrowdedPrep
from sfft.utils.meta.TimeoutKit import TimeoutAfter
from sfft.utils.SFFTSolutionReader import Realize_FluxScaling
from sfft.sfftcore.SFFTSubtract import GeneralSFFTSubtract
from sfft.sfftcore.SFFTConfigure import SingleSFFTConfigure
# version: Feb 24, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class MultiEasy_CrowdedPacket:
    def __init__(self, FITS_REF_Queue, FITS_SCI_Queue, FITS_DIFF_Queue=[], FITS_Solution_Queue=[], \
        ForceConv_Queue=[], GKerHW_Queue=[], KerHWRatio=2.0, KerHWLimit=(2, 20), KerPolyOrder=2, \
        BGPolyOrder=2, ConstPhotRatio=True, MaskSatContam=False, GAIN_KEY='GAIN', SATUR_KEY='SATURATE', \
        BACK_TYPE='AUTO', BACK_VALUE='0.0', BACK_SIZE=64, BACK_FILTERSIZE=3, DETECT_THRESH=5.0, \
        ANALYSIS_THRESH=5.0, DETECT_MINAREA=5, DETECT_MAXAREA=0, DEBLEND_MINCONT=0.005, BACKPHOTO_TYPE='LOCAL', \
        ONLY_FLAGS=None, BoundarySIZE=0.0, BACK_SIZE_SUPER=128, StarExt_iter=2, PriorBanMask_Queue=[], \
        CLEAN_GPU_MEMORY=False, VERBOSE_LEVEL=2):

        """
        # NOTE: This function is to perform Crowded-Flavor SFFT for multiple tasks:
        #       @ this module (only support Cupy backend) would help making the 
        #         best use of the available CPU and GPU resources.

        * Parameters for Crowded-Flavor SFFT [multiple tasks]
        
        # ----------------------------- Computing Enviornment [Cupy backend ONLY] --------------------------------- #

        -NUM_THREADS_4PREPROC [8]           # the number of Python threads for preprocessing.
                                            # WARNING: Empirically, the current code is only optimized for 
                                            #          NUM_TASK / NUM_THREADS_4PREPROC in [1.0 - 4.0].
                                            #          The computing effeciency goes down sharply as the ratio increases.
                                            #          Too large number of tasks may even cause memory overflow.

        -NUM_THREADS_4SUBTRACT [1]          # the number of Python threads for sfft subtraction.
                                            # says, if you have four GPU cards, then set it to 4.
                                            # NOTE: please do not confused the Python threads 
                                            #       with the CUDA threads (block, threads ...).

        -CUDA_DEVICES_4SUBTRACT [['0']]     # it specifies certain GPU devices (indices) to conduct the subtractions.
                                            # the GPU devices are usually numbered 0 to N-1. 
                                            # (you may use command nvidia-smi to check for Nvidia GPU devices).

        -TIMEOUT_4PREPROC_EACHTASK [300]    # timeout of each task during preprocessing.
                                            
        -TIMEOUT_4SUBTRACT_EACHTASK [300]   # timeout of each task during sfft subtraction.

        # ----------------------------- Preprocessing with Saturation Rejection for Image-Masking --------------------------------- #

        # > Configurations for SExtractor

        -GAIN_KEY ['GAIN']                  # SExtractor Parameter GAIN_KEY
                                            # i.e., keyword of GAIN in FITS header (of reference & science)

        -SATUR_KEY ['SATURATE']             # SExtractor Parameter SATUR_KEY
                                            # i.e., keyword of effective saturation in FITS header (of reference & science)
                                            # Remarks: note that Crowded-Flavor SFFT does not require sky-subtracted images as inputs,
                                            #          so the default keyword is the common name for saturation level.
        
        -BACK_TYPE ['AUTO']                 # SExtractor Parameter BACK_TYPE = [AUTO or MANUAL].
         
        -BACK_VALUE [0]                     # SExtractor Parameter BACK_VALUE (only work for BACK_TYPE='MANUAL')

        -BACK_SIZE [64]                     # SExtractor Parameter BACK_SIZE

        -BACK_FILTERSIZE [3]                # SExtractor Parameter BACK_FILTERSIZE

        -DETECT_THRESH [5.0]                # SExtractor Parameter DETECT_THRESH
                                            # NOTE: One may notice that the default DETECT_THRESH in SExtractor is 1.5.
                                            #       Using a 'very cold' detection threshold here is to speed up SExtractor.
                                            #       Although DETECT_THRESH = 5.0 means we will miss the faint-end sources with SNR < 20 
                                            #       (approximately), the cost is generally acceptable for saturation mask.

        -ANALYSIS_THRESH [5.0]              # SExtractor Parameter ANALYSIS_THRESH
                                            # NOTE: By default, let -ANALYSIS_THRESH = -DETECT_THRESH

        -DETECT_MINAREA [5]                 # SExtractor Parameter DETECT_MINAREA
        
        -DETECT_MAXAREA [0]                 # SExtractor Parameter DETECT_MAXAREA

        -DEBLEND_MINCONT [0.005]            # SExtractor Parameter DEBLEND_MINCONT (typically, 0.001 - 0.005)

        -BACKPHOTO_TYPE ['LOCAL']           # SExtractor Parameter BACKPHOTO_TYPE

        -ONLY_FLAGS [None]                  # Restrict SExtractor Output Photometry Catalog by Source FLAGS
                                            # Common FLAGS (Here None means no restrictions on FLAGS):
                                            # 1: aperture photometry is likely to be biased by neighboring sources 
                                            #    or by more than 10% of bad pixels in any aperture
                                            # 2: the object has been deblended
                                            # 4: at least one object pixel is saturated

        -BoundarySIZE [30]                  # Restrict SExtractor Output Photometry Catalog by Dropping Sources at Boundary 
                                            # NOTE: This would help to avoid selecting sources too close to image boundary. 

        # > Other Configurations

        -BACK_SIZE_SUPER [128]              # Also a SExtractor configuration BACK_SIZE, however, the background is 
                                            # used to fill the masked regions (saturation contaminated pixels).
                                            # NOTE: '_SUPER' means we would like to use a realtively large BACK_SIZE to
                                            #       make a super-smooth (structure-less) background. So the default
                                            #       -BACK_SIZE_SUPER > -BACK_SIZE

        -StarExt_iter [2]                   # make a further dilation for the masked region initially determined by SExtractor SEGMENTATION.
                                            # -StarExt_iter means the iteration times of the dilation process. 
        
        -PriorBanMask_Queue [[]]]           # A queue of -PriorBanMask in sfft.Easy_CrowdedPacket.
                                            # Recall -PriorBanMask:
                                            # a Numpy boolean array, with shape consistent with reference (science).
                                            # one can deliver a customized mask that covers known 
                                            # variables/transients/bad-pixels by this argument.
                                            # SFFT will then merge (Union operation) the user-defined mask and 
                                            # the SExtractor-determined saturation mask as the final mask.

        # ----------------------------- SFFT Subtraction --------------------------------- #

        -ForceConv_Queue ['AUTO']           # A queue of -ForceConv in sfft.Easy_CrowdedPacket.
                                            # it determines which image will be convolved, can be 'REF', 'SCI' and 'AUTO'.
                                            # -ForceConv = 'AUTO' means SFFT will determine the convolution direction according to 
                                            # FWHM_SCI and FWHM_REF: the image with better seeing will be convolved to avoid deconvolution.

        -GKerHW_Queue [None]                # A queue of -GKerHW in sfft.Easy_CrowdedPacket.
                                            # The given kernel half-width, None means the kernel size will be 
                                            # automatically determined by -KerHWRatio (to be seeing-related). 

        -KerHWRatio [2.0]                   # The ratio between FWHM and the kernel half-width
                                            # KerHW = int(KerHWRatio * Max(FWHM_REF, FWHM_SCI))

        -KerHWLimit [(2, 20)]               # The lower & upper bounds for kernel half-width 
                                            # KerHW is updated as np.clip(KerHW, KerHWLimit[0], KerHWLimit[1]) 
                                            # Remarks: this is useful for a survey since it can constrain the peak GPU memory usage.

        -KerPolyOrder [2]                   # Polynomial degree of kernel spatial variation, can be [0,1,2,3].

        -BGPolyOrder [2]                    # Polynomial degree of differential-background spatial variation, can be [0,1,2,3].
                                            # It is non-trivial for Crowded-Flavor SFFT, as the input images are usually not sky subtracted.
        
        -ConstPhotRatio [True]              # Constant photometric ratio between images? can be True or False
                                            # ConstPhotRatio = True: the sum of convolution kernel is restricted 
                                            #   to be a constant across the field. 
                                            # ConstPhotRatio = False: the flux scaling between images is modeled
                                            #   by a polynomial with degree -KerPolyOrder.

        -MaskSatContam [False]              # Mask saturation-contaminated regions on difference image ? can be True or False
                                            # NOTE the pixels enclosed in the regions are replaced by NaN.

        # ----------------------------- Input & Output --------------------------------- #

        -FITS_REF_Queue []                  # A queue of -FITS_REF in sfft.Easy_CrowdedPacket.
                                            # File path of input reference image.

        -FITS_SCI_Queue []                  # A queue of -FITS_SCI in sfft.Easy_CrowdedPacket.
                                            # File path of input science image.

        -FITS_DIFF_Queue [None]             # A queue of -FITS_DIFF in sfft.Easy_CrowdedPacket.
                                            # File path of output difference image.

        -FITS_Solution_Queue [None]         # A queue of -FITS_Solution in sfft.Easy_CrowdedPacket.
                                            # File path of the solution of the linear system.
                                            # it is an array of (..., a_ijab, ... b_pq, ...).

        # ----------------------------- Miscellaneous --------------------------------- #
        
        -CLEAN_GPU_MEMORY [False]           # Clean GPU memory or not after all subtractions done in a GPU device.

        -VERBOSE_LEVEL [2]                  # The level of verbosity, can be [0, 1, 2]
                                            # 0/1/2: QUIET/NORMAL/FULL mode

        # Important Notice:
        #
        # a): if reference is convolved in SFFT, then DIFF = SCI - Convolved_REF.
        #     [difference image is expected to have PSF & flux zero-point consistent with science image]
        #     e.g., -ForceConv='REF' or -ForceConv='AUTO' when reference has better seeing.
        #
        # b): if science is convolved in SFFT, then DIFF = Convolved_SCI - REF
        #     [difference image is expected to have PSF & flux zero-point consistent with reference image]
        #     e.g., -ForceConv='SCI' or -ForceConv='AUTO' when science has better seeing.
        #
        # Remarks: this convention is to guarantee that transients emerge on science image always 
        #          show a positive signal on difference images.

        """

        self.FITS_REF_Queue = FITS_REF_Queue
        self.FITS_SCI_Queue = FITS_SCI_Queue
        self.NUM_TASK = len(FITS_SCI_Queue)

        self.KerHWRatio = KerHWRatio
        self.KerHWLimit = KerHWLimit
        self.KerPolyOrder = KerPolyOrder
        self.BGPolyOrder = BGPolyOrder
        self.ConstPhotRatio = ConstPhotRatio
        self.MaskSatContam = MaskSatContam
        
        self.GAIN_KEY = GAIN_KEY
        self.SATUR_KEY = SATUR_KEY
        self.BACK_TYPE = BACK_TYPE
        self.BACK_VALUE = BACK_VALUE
        self.BACK_SIZE = BACK_SIZE
        self.BACK_FILTERSIZE = BACK_FILTERSIZE
        
        self.DETECT_THRESH = DETECT_THRESH
        self.ANALYSIS_THRESH = ANALYSIS_THRESH
        self.DETECT_MINAREA = DETECT_MINAREA
        self.DETECT_MAXAREA = DETECT_MAXAREA
        self.DEBLEND_MINCONT = DEBLEND_MINCONT
        self.BACKPHOTO_TYPE = BACKPHOTO_TYPE
        self.ONLY_FLAGS = ONLY_FLAGS
        self.BoundarySIZE = BoundarySIZE
        
        self.BACK_SIZE_SUPER = BACK_SIZE_SUPER
        self.StarExt_iter = StarExt_iter
        
        self.CLEAN_GPU_MEMORY = CLEAN_GPU_MEMORY
        self.VERBOSE_LEVEL = VERBOSE_LEVEL

        if FITS_DIFF_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument FITS_DIFF_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.FITS_DIFF_Queue = [None] * self.NUM_TASK
        else: self.FITS_DIFF_Queue = FITS_DIFF_Queue
        
        if FITS_Solution_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument FITS_Solution_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.FITS_Solution_Queue = [None] * self.NUM_TASK
        else: self.FITS_Solution_Queue = FITS_Solution_Queue
        
        if ForceConv_Queue == []:
            if self.VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = 'Argument ForceConv_Queue is EMPTY then Use default [...AUTO...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.ForceConv_Queue = ['AUTO'] * self.NUM_TASK
        else: self.ForceConv_Queue = ForceConv_Queue

        if GKerHW_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument GKerHW_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.GKerHW_Queue = [None] * self.NUM_TASK
        else: self.GKerHW_Queue = GKerHW_Queue

        if PriorBanMask_Queue == []:
            if self.VERBOSE_LEVEL in [2]:
                _warn_message = 'Argument PriorBanMask_Queue is EMPTY then Use default [...None...] instead!'
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)
            self.PriorBanMask_Queue = [None] * self.NUM_TASK
        else: self.PriorBanMask_Queue = PriorBanMask_Queue

    def MESP_Cupy(self, NUM_THREADS_4PREPROC=8, NUM_THREADS_4SUBTRACT=1, CUDA_DEVICES_4SUBTRACT=['0'], \
        TIMEOUT_4PREPROC_EACHTASK=300, TIMEOUT_4SUBTRACT_EACHTASK=300):

        import cupy as cp
        
        # * define task-oriented dictionaries of status-bar and products
        #   status 0: Initial State
        #   status 1: Preprocessing SUCCESS || status -1: Preprocessing FAIL
        #   status 2: Subtraction SUCCESS || status -2: Subtraction FAIL
        #   status 32: Preprocessing is IN-PROGRESS ...
        #   status 64: Subtraction is IN-PROGRESS ...
        #   NOTE: Subtraction would not be triggered when Preprocessing fails
        
        _RATIO = self.NUM_TASK / NUM_THREADS_4PREPROC
        if _RATIO > 4.0:
            if self.VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = 'THIS FUNCTION IS NOT OPTIMIZED FOR RATIO OF '
                _warn_message += 'NUM_TASK / NUM_THREADS_4PREPROC BEING TO HIGH [%.1f > 4.0]!' %_RATIO
                warnings.warn('\nMeLOn WARNING: %s' %_warn_message)

        BACKEND_4SUBTRACT = 'Cupy'
        taskidx_lst = np.arange(self.NUM_TASK)
        assert NUM_THREADS_4SUBTRACT == len(CUDA_DEVICES_4SUBTRACT)
        
        DICT_STATUS_BAR = {'TASK-[%d]' %taskidx: 0 for taskidx in taskidx_lst}
        DICT_PRODUCTS = {'TASK-[%d]' %taskidx: {} for taskidx in taskidx_lst}
        lock = threading.RLock()

        # * define the function for preprocessing (see sfft.EasyCrowdedPacket)
        def FUNC_4PREPROC(INDEX_THREAD_4PREPROC):

            THREAD_ALIVE_4PREPROC = True
            while THREAD_ALIVE_4PREPROC:
                with lock:
                    STATUS_SNAPSHOT = np.array([DICT_STATUS_BAR['TASK-[%d]' %taskidx] for taskidx in taskidx_lst])
                    OPEN_MASK_4PREPROC = STATUS_SNAPSHOT == 0
                    if np.sum(OPEN_MASK_4PREPROC) > 0:
                        taskidx_acquired = taskidx_lst[OPEN_MASK_4PREPROC][0]
                        DICT_STATUS_BAR['TASK-[%d]' %taskidx_acquired] = 32

                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4PREPROC-[%d] ACQUIRED TASK-[%d] ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                            _message += '(Crowded-Flavor Auto Preprocessing)!'
                            print('\nMeLOn CheckPoint: %s' %_message)
                    else: 
                        THREAD_ALIVE_4PREPROC = False
                        taskidx_acquired = None

                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4PREPROC-[%d] TERMINATED ' %INDEX_THREAD_4PREPROC
                            _message += 'SINCE NO MORE TASK WILL BE ALLOCATED '
                            _message += '(Crowded-Flavor Auto Preprocessing)!'
                            print('\nMeLOn CheckPoint: %s' %_message)

                if taskidx_acquired is not None:
                    FITS_REF = self.FITS_REF_Queue[taskidx_acquired]
                    FITS_SCI = self.FITS_SCI_Queue[taskidx_acquired]
                    PriorBanMask = self.PriorBanMask_Queue[taskidx_acquired]

                    try:
                        with TimeoutAfter(timeout=TIMEOUT_4PREPROC_EACHTASK, exception=TimeoutError):
                            
                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4PREPROC-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                                _message += 'TRIGGER Crowded-Flavor Auto Preprocessing!'
                                print('MeLOn CheckPoint: %s' %_message)

                            # ** perform Auto-Crowded-Prep
                            _ACP = Auto_CrowdedPrep(FITS_REF=FITS_REF, FITS_SCI=FITS_SCI, GAIN_KEY=self.GAIN_KEY, \
                                SATUR_KEY=self.SATUR_KEY, BACK_TYPE=self.BACK_TYPE, BACK_VALUE=self.BACK_VALUE, \
                                BACK_SIZE=self.BACK_SIZE, BACK_FILTERSIZE=self.BACK_FILTERSIZE, \
                                DETECT_THRESH=self.DETECT_THRESH, ANALYSIS_THRESH=self.ANALYSIS_THRESH, \
                                DETECT_MINAREA=self.DETECT_MINAREA, DETECT_MAXAREA=self.DETECT_MAXAREA, \
                                DEBLEND_MINCONT=self.DEBLEND_MINCONT, BACKPHOTO_TYPE=self.BACKPHOTO_TYPE, \
                                ONLY_FLAGS=self.ONLY_FLAGS, BoundarySIZE=self.BoundarySIZE, VERBOSE_LEVEL=self.VERBOSE_LEVEL)
                            
                            SFFTPrepDict = _ACP.AutoMask(BACK_SIZE_SUPER=self.BACK_SIZE_SUPER, \
                                StarExt_iter=self.StarExt_iter, PriorBanMask=PriorBanMask)
                        
                        NEW_STATUS = 1
                        if self.VERBOSE_LEVEL in [1, 2]:
                            _message = 'THREAD-4PREPROC-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                            _message += 'SUCCESSFUL Crowded-Flavor Auto Preprocessing!'
                            print('\nMeLOn CheckPoint: %s' %_message)

                    except Exception as e:
                        NEW_STATUS = -1
                        SFFTPrepDict = None
                        if self.VERBOSE_LEVEL in [0, 1, 2]:
                            print(e, file=sys.stderr)
                            _error_message = 'THREAD-4PREPROC-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4PREPROC, taskidx_acquired)
                            _error_message += 'FAILED Crowded-Flavor Auto Preprocessing!'
                            print('\nMeLOn ERROR: %s' %_error_message)
                    
                    with lock:
                        # NOTE: IN FACT, THERE IS NO RISK HERE EVEN WITHOUT LOCK.
                        DICT_STATUS_BAR['TASK-[%d]' %taskidx_acquired] = NEW_STATUS
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['SFFTPrepDict'] = SFFTPrepDict

            return None
        
        # * define the function for sfft subtraction (see sfft.EasyCrowdedPacket)
        def FUNC_4SUBTRCT(INDEX_THREAD_4SUBTRACT):
            
            CUDA_DEVICE_4SUBTRACT = CUDA_DEVICES_4SUBTRACT[INDEX_THREAD_4SUBTRACT]
            device = cp.cuda.Device(int(CUDA_DEVICE_4SUBTRACT))
            device.use()

            THREAD_ALIVE_4SUBTRCT = True
            GPU_MEMORY_CLEANED = True

            while THREAD_ALIVE_4SUBTRCT:
                SHORT_PAUSE = False

                with lock:
                    STATUS_SNAPSHOT = np.array([DICT_STATUS_BAR['TASK-[%d]' %taskidx] for taskidx in taskidx_lst])
                    OPEN_MASK_4SUBTRACT = STATUS_SNAPSHOT == 1
                    WAIT_MASK_4SUBTRACT = np.in1d(STATUS_SNAPSHOT, [0, 32])

                    if np.sum(OPEN_MASK_4SUBTRACT) > 0:
                        taskidx_acquired = taskidx_lst[OPEN_MASK_4SUBTRACT][0]
                        DICT_STATUS_BAR['TASK-[%d]' %taskidx_acquired] = 64

                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4SUBTRACT-[%d] ACQUIRED TASK-[%d] ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                            _message += '(SFFT-SUBTRACTION)!'
                            print('\nMeLOn CheckPoint: %s' %_message)

                    elif np.sum(WAIT_MASK_4SUBTRACT) > 0:
                        SHORT_PAUSE = True
                        taskidx_acquired = None

                    else:
                        THREAD_ALIVE_4SUBTRCT = False
                        taskidx_acquired = None

                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4SUBTRACT-[%d] TERMINATED ' %INDEX_THREAD_4SUBTRACT
                            _message += 'SINCE NO MORE TASK WILL BE ALLOCATED '
                            _message += '(SFFT-SUBTRACTION)!'
                            print('\nMeLOn CheckPoint: %s' %_message)

                if SHORT_PAUSE:
                    # run a short sleep, otherwise this thread may always hold the lock by the while loop
                    time.sleep(0.01)
                
                if taskidx_acquired is not None:

                    FITS_REF = self.FITS_REF_Queue[taskidx_acquired]
                    FITS_SCI = self.FITS_SCI_Queue[taskidx_acquired]
                    FITS_DIFF = self.FITS_DIFF_Queue[taskidx_acquired]
                    FITS_Solution = self.FITS_Solution_Queue[taskidx_acquired]
                    ForceConv = self.ForceConv_Queue[taskidx_acquired]
                    GKerHW = self.GKerHW_Queue[taskidx_acquired]

                    with lock:
                        # NOTE: IN FACT, THERE IS NO RISK HERE EVEN WITHOUT LOCK.
                        SFFTPrepDict = DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['SFFTPrepDict']

                    try:
                        with TimeoutAfter(timeout=TIMEOUT_4SUBTRACT_EACHTASK, exception=TimeoutError):

                            # * Determine ConvdSide & KerHW
                            FWHM_REF = SFFTPrepDict['FWHM_REF']
                            FWHM_SCI = SFFTPrepDict['FWHM_SCI']

                            assert ForceConv in ['AUTO', 'REF', 'SCI']
                            if ForceConv == 'AUTO':
                                if FWHM_SCI >= FWHM_REF: ConvdSide = 'REF'
                                else: ConvdSide = 'SCI'
                            else: ConvdSide = ForceConv

                            if GKerHW is None:
                                FWHM_La = np.max([FWHM_REF, FWHM_SCI])
                                KerHW = int(np.clip(self.KerHWRatio * FWHM_La, \
                                                    self.KerHWLimit[0], self.KerHWLimit[1]))
                            else: KerHW = GKerHW

                            # * Compile Functions in SFFT Subtraction
                            PixA_REF = SFFTPrepDict['PixA_REF']
                            PixA_SCI = SFFTPrepDict['PixA_SCI']
                            
                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'TRIGGER Function Compilations of SFFT-SUBTRACTION'
                                print('MeLOn CheckPoint: %s' %_message)
                            
                            Tcomp_start = time.time()
                            GPU_MEMORY_CLEANED = False
                            SFFTConfig = SingleSFFTConfigure.SSC(NX=PixA_REF.shape[0], NY=PixA_REF.shape[1], \
                                KerHW=KerHW, KerPolyOrder=self.KerPolyOrder, BGPolyOrder=self.BGPolyOrder, \
                                ConstPhotRatio=self.ConstPhotRatio, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
                                VERBOSE_LEVEL=self.VERBOSE_LEVEL)

                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'Function Compilations of SFFT-SUBTRACTION '
                                _message += 'TAKES [%.3f s]!' %(time.time() - Tcomp_start)
                                print('\nMeLOn Report: %s' %_message)

                            # * Perform SFFT Subtraction
                            SatMask_REF = SFFTPrepDict['REF-SAT-Mask']
                            SatMask_SCI = SFFTPrepDict['SCI-SAT-Mask']
                            NaNmask_U = SFFTPrepDict['Union-NaN-Mask']
                            PixA_mREF = SFFTPrepDict['PixA_mREF']
                            PixA_mSCI = SFFTPrepDict['PixA_mSCI']

                            if ConvdSide == 'REF':
                                PixA_mI, PixA_mJ = PixA_mREF, PixA_mSCI
                                if NaNmask_U is not None:
                                    PixA_I, PixA_J = PixA_REF.copy(), PixA_SCI.copy()
                                    PixA_I[NaNmask_U] = PixA_mI[NaNmask_U]
                                    PixA_J[NaNmask_U] = PixA_mJ[NaNmask_U]
                                else: PixA_I, PixA_J = PixA_REF, PixA_SCI
                                if self.MaskSatContam: 
                                    ContamMask_I = SatMask_REF
                                    ContamMask_J = SatMask_SCI
                                else: ContamMask_I = None

                            if ConvdSide == 'SCI':
                                PixA_mI, PixA_mJ = PixA_mSCI, PixA_mREF
                                if NaNmask_U is not None:
                                    PixA_I, PixA_J = PixA_SCI.copy(), PixA_REF.copy()
                                    PixA_I[NaNmask_U] = PixA_mI[NaNmask_U]
                                    PixA_J[NaNmask_U] = PixA_mJ[NaNmask_U]
                                else: PixA_I, PixA_J = PixA_SCI, PixA_REF
                                if self.MaskSatContam: 
                                    ContamMask_I = SatMask_SCI
                                    ContamMask_J = SatMask_REF
                                else: ContamMask_I = None
                            
                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'TRIGGER Function Executions of SFFT-SUBTRACTION!'
                                print('MeLOn CheckPoint: %s' %_message)
                            
                            Tsub_start = time.time()
                            _tmp = GeneralSFFTSubtract.GSS(PixA_I=PixA_I, PixA_J=PixA_J, PixA_mI=PixA_mI, PixA_mJ=PixA_mJ, \
                                SFFTConfig=SFFTConfig, ContamMask_I=ContamMask_I, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
                                VERBOSE_LEVEL=self.VERBOSE_LEVEL)

                            Solution, PixA_DIFF, ContamMask_CI = _tmp
                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'Function Executions of SFFT-SUBTRACTION '
                                _message += 'TAKES [%.3f s]!' %(time.time() - Tsub_start)
                                print('\nMeLOn Report: %s' %_message)

                            # ** adjust the sign of difference image 
                            #    NOTE: Our convention is to make the transients on SCI always show themselves as positive signal on DIFF
                            #    (a) when REF is convolved, DIFF = SCI - Conv(REF)
                            #    (b) when SCI is convolved, DIFF = Conv(SCI) - REF

                            if ConvdSide == 'SCI':
                                PixA_DIFF = -PixA_DIFF

                            if self.VERBOSE_LEVEL in [1, 2]:
                                if ConvdSide == 'REF':
                                    _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                    _message += 'Reference Image is Convolved in SFFT-SUBTRACTION [DIFF = SCI - Conv(REF)]!'
                                    print('MeLOn CheckPoint: %s' %_message)

                                if ConvdSide == 'SCI':
                                    _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                    _message += 'Science Image is Convolved in SFFT-SUBTRACTION [DIFF = Conv(SCI) - REF]!'
                                    print('MeLOn CheckPoint: %s' %_message)

                            # ** estimate the flux scaling through the convolution of image subtraction
                            #    NOTE: if photometric ratio is not constant (ConstPhotRatio=False), we measure of a grid
                            #          of coordinates to estimate the flux scaling and its fluctuation (polynomial form).
                            #    NOTE: here we set the tile-size of the grid about 64 x 64 pix, but meanwhile, 
                            #          we also require the tiles should no less than 6 along each axis.

                            N0, N1 = SFFTConfig[0]['N0'], SFFTConfig[0]['N1']
                            L0, L1 = SFFTConfig[0]['L0'], SFFTConfig[0]['L1']
                            DK, Fpq = SFFTConfig[0]['DK'], SFFTConfig[0]['Fpq']

                            if self.ConstPhotRatio:
                                SFFT_FSCAL_NSAMP = 1
                                XY_q = np.array([[N0/2.0, N1/2.0]]) + 0.5
                                RFS = Realize_FluxScaling(XY_q=XY_q)
                                SFFT_FSCAL_q = RFS.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)
                                SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG = SFFT_FSCAL_q[0], 0.0
                            
                            if not self.ConstPhotRatio:
                                TILESIZE_X, TILESIZE_Y = 64, 64    # FIXME emperical/arbitrary values
                                NTILE_X = np.max([round(N0/TILESIZE_X), 6])
                                NTILE_Y = np.max([round(N1/TILESIZE_Y), 6])
                                GX = np.linspace(0.5, N0+0.5, NTILE_X+1)
                                GY = np.linspace(0.5, N1+0.5, NTILE_Y+1)
                                YY, XX = np.meshgrid(GY, GX)
                                XY_q = np.array([XX.flatten(), YY.flatten()]).T
                                SFFT_FSCAL_NSAMP = XY_q.shape[0]
                                
                                RFS = Realize_FluxScaling(XY_q=XY_q)
                                SFFT_FSCAL_q = RFS.FromArray(Solution=Solution, N0=N0, N1=N1, L0=L0, L1=L1, DK=DK, Fpq=Fpq)
                                SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG = np.mean(SFFT_FSCAL_q), np.std(SFFT_FSCAL_q)

                            if self.VERBOSE_LEVEL in [1, 2]:
                                _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                                _message += 'The Flux Scaling through the Convolution of SFFT-SUBTRACTION '
                                _message += '[%.6f +/- %.6f] from [%d] positions!\n' %(SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG, SFFT_FSCAL_NSAMP)
                                print('MeLOn CheckPoint: %s' %_message)

                            # ** final tweaks on the difference image
                            if NaNmask_U is not None:
                                PixA_DIFF[NaNmask_U] = np.nan   # Mask Union-NaN regions
                            
                            if self.MaskSatContam:
                                ContamMask_DIFF = np.logical_or(ContamMask_CI, ContamMask_J)
                                PixA_DIFF[ContamMask_DIFF] = np.nan   # Mask Saturation-Contaminated regions
                            
                            # * Save difference image
                            if FITS_DIFF is not None:

                                """
                                # Remarks on header of difference image
                                # [1] In general, the FITS header of difference image would mostly be inherited from science image.
                                #
                                # [2] when science image is convolved, we turn to set GAIN_DIFF = GAIN_SCI / SFFT_FSCAL_MEAN (unit: e-/ADU)
                                #     so that one can correctly convert ADU to e- for a transient appears on science image.
                                #     WARNING: for a transient appears on reference image, GAIN_DIFF = GAIN_SCI is correct.
                                #              we should keep in mind that GAIN_DIFF is not an absolute instrumental value.
                                #     WARNING: one can only estimate the Poission noise from the science transient via GIAN_DIFF.
                                #              the Poission noise from the background source (host galaxy) can enhance the 
                                #              background noise at the position of the transient. photometry softwares which 
                                #              only take background noise and transient Possion noise into account would tend 
                                #              to overestimate the SNR of the transient.
                                # 
                                # [3] when science image is convolved, we turn to set SATURA_DIFF = SATURA_SCI * SFFT_FSCAL_MEAN         
                                #     WARNING: it is still not a good idea to use SATURA_DIFF to identify the SATURATION regions
                                #              on difference image. more appropriate way is masking the saturation contaminated 
                                #              regions on difference image (MaskSatContam=True), or alternatively, leave them  
                                #              alone and using AI stamp classifier to reject saturation-related bogus.
                                #
                                """

                                _hdl = fits.open(FITS_SCI)
                                _hdl[0].data[:, :] = PixA_DIFF.T

                                _hdl[0].header['NAME_REF'] = (pa.basename(FITS_REF), 'MeLOn: SFFT')
                                _hdl[0].header['NAME_SCI'] = (pa.basename(FITS_SCI), 'MeLOn: SFFT')
                                _hdl[0].header['FWHM_REF'] = (FWHM_REF, 'MeLOn: SFFT')
                                _hdl[0].header['FWHM_SCI'] = (FWHM_SCI, 'MeLOn: SFFT')
                                _hdl[0].header['KERORDER'] = (self.KerPolyOrder, 'MeLOn: SFFT')
                                _hdl[0].header['BGORDER'] = (self.BGPolyOrder, 'MeLOn: SFFT')
                                _hdl[0].header['CPHOTR'] = (str(self.ConstPhotRatio), 'MeLOn: SFFT')
                                _hdl[0].header['KERHW'] = (KerHW, 'MeLOn: SFFT')
                                _hdl[0].header['CONVD'] = (ConvdSide, 'MeLOn: SFFT')

                                if ConvdSide == 'SCI':
                                    GAIN_SCI = _hdl[0].header[self.GAIN_KEY]
                                    SATUR_SCI = _hdl[0].header[self.SATUR_KEY]
                                    GAIN_DIFF = GAIN_SCI / SFFT_FSCAL_MEAN
                                    SATUR_DIFF = SATUR_SCI * SFFT_FSCAL_MEAN

                                    _hdl[0].header[self.GAIN_KEY] = (GAIN_DIFF, 'MeLOn: SFFT')
                                    _hdl[0].header[self.SATUR_KEY] = (SATUR_DIFF, 'MeLOn: SFFT')

                                _hdl.writeto(FITS_DIFF, overwrite=True)
                                _hdl.close()
                            
                            # * Save solution array
                            if FITS_Solution is not None:
                                phdu = fits.PrimaryHDU()
                                phdu.header['N0'] = (SFFTConfig[0]['N0'], 'MeLOn: SFFT')
                                phdu.header['N1'] = (SFFTConfig[0]['N1'], 'MeLOn: SFFT')
                                phdu.header['DK'] = (SFFTConfig[0]['DK'], 'MeLOn: SFFT')
                                phdu.header['DB'] = (SFFTConfig[0]['DB'], 'MeLOn: SFFT')
                                phdu.header['L0'] = (SFFTConfig[0]['L0'], 'MeLOn: SFFT')
                                phdu.header['L1'] = (SFFTConfig[0]['L1'], 'MeLOn: SFFT')
                                
                                phdu.header['FIJ'] = (SFFTConfig[0]['Fij'], 'MeLOn: SFFT')
                                phdu.header['FAB'] = (SFFTConfig[0]['Fab'], 'MeLOn: SFFT')
                                phdu.header['FPQ'] = (SFFTConfig[0]['Fpq'], 'MeLOn: SFFT')
                                phdu.header['FIJAB'] = (SFFTConfig[0]['Fijab'], 'MeLOn: SFFT')

                                PixA_Solution = Solution.reshape((-1, 1))
                                phdu.data = PixA_Solution.T
                                fits.HDUList([phdu]).writeto(FITS_Solution, overwrite=True)

                        NEW_STATUS = 2
                        if self.VERBOSE_LEVEL in [1, 2]:
                            _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                            _message += 'SUCCESSFUL SFFT-SUBTRACTION!'
                            print('\nMeLOn CheckPoint: %s' %_message)
                    
                    except Exception as e:
                        PixA_DIFF, Solution = None, None
                        SFFT_FSCAL_MEAN, SFFT_FSCAL_SIG = np.nan, np.nan

                        # ** clean GPU memory when the subtraction fails
                        mempool = cp.get_default_memory_pool()
                        pinned_mempool = cp.get_default_pinned_memory_pool()
                        mempool.free_all_blocks()
                        pinned_mempool.free_all_blocks()
                        
                        GPU_MEMORY_CLEANED = True
                        if self.VERBOSE_LEVEL in [2]:
                            _message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                            _message += 'CLEAN GPU-[%s] MEMORY!' %CUDA_DEVICE_4SUBTRACT
                            print('\nMeLOn CheckPoint: %s' %_message)

                        NEW_STATUS = -2
                        Solution = None
                        PixA_DIFF = None
                        if self.VERBOSE_LEVEL in [0, 1, 2]:
                            print(e, file=sys.stderr)
                            _error_message = 'THREAD-4SUBTRACT-[%d] & TASK-[%d]: ' %(INDEX_THREAD_4SUBTRACT, taskidx_acquired)
                            _error_message += 'FAILED SFFT-SUBTRACTION!'
                            print('\nMeLOn ERROR: %s' %_error_message)

                    with lock:
                        DICT_STATUS_BAR['TASK-[%d]' %taskidx_acquired] = NEW_STATUS
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['PixA_DIFF'] = PixA_DIFF
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['Solution'] = Solution
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['SFFT_FSCAL_MEAN'] = SFFT_FSCAL_MEAN
                        DICT_PRODUCTS['TASK-[%d]' %taskidx_acquired]['SFFT_FSCAL_SIG'] = SFFT_FSCAL_SIG

            if self.CLEAN_GPU_MEMORY:
                if not GPU_MEMORY_CLEANED:
                    mempool = cp.get_default_memory_pool()
                    pinned_mempool = cp.get_default_pinned_memory_pool()
                    mempool.free_all_blocks()
                    pinned_mempool.free_all_blocks()
                    
                    if self.VERBOSE_LEVEL in [1, 2]:
                        _message = 'THREAD-4SUBTRACT-[%d]: ' %INDEX_THREAD_4SUBTRACT
                        _message += 'FINAL CLEAN GPU-[%s] MEMORY ' %CUDA_DEVICE_4SUBTRACT
                        print('\nMeLOn CheckPoint: %s' %_message)
                else:
                    if self.VERBOSE_LEVEL in [1, 2]:
                        _message = 'THREAD-4SUBTRACT-[%d]: ' %INDEX_THREAD_4SUBTRACT
                        _message += 'SKIP FINAL CLEAN, SINCE GPU-[%s] MEMORY ALREADY CLEANED' %CUDA_DEVICE_4SUBTRACT
                        print('\nMeLOn CheckPoint: %s' %_message)

            return None

        # * trigger the threads for the preprocessing and subtraction
        MyThreadQueue = []
        for INDEX_THREAD_4PREPROC in range(NUM_THREADS_4PREPROC):
            MyThread = threading.Thread(target=FUNC_4PREPROC, args=(INDEX_THREAD_4PREPROC,))
            MyThreadQueue.append(MyThread)
            MyThread.start()

        for INDEX_THREAD_4SUBTRACT in range(NUM_THREADS_4SUBTRACT):
            MyThread = threading.Thread(target=FUNC_4SUBTRCT, args=(INDEX_THREAD_4SUBTRACT,))
            MyThreadQueue.append(MyThread)
            MyThread.start()
  
        for MyThread in MyThreadQueue:
            MyThread.join()
        
        # * show a summary
        if self.VERBOSE_LEVEL in [0, 1, 2]:
            NUM_SUCCESS = np.sum(np.array([DICT_STATUS_BAR[task] for task in DICT_STATUS_BAR]) == 2)
            print('\nMeLOn CheckPoint: SFFT PROCESSED TASKS SUCCESSFULLY [%d / %d]!' %(NUM_SUCCESS, self.NUM_TASK))
        
        return DICT_STATUS_BAR, DICT_PRODUCTS



"""
Remarks on Internal Packages Imports:
    NONE
    
"""

from .SFFTConfigure import SingleSFFTConfigure
from .SFFTSubtract import ElementalSFFTSubtract, GeneralSFFTSubtract, GeneralSFFTSubtract_PureCupy



import bisect
import skimage
import warnings
import numpy as np
from skimage import feature
from pkg_resources import parse_version
from skimage.transform import hough_line, hough_line_peaks
# version: Feb 4, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Hough_Detection:
    @staticmethod
    def HD(XY_obj=None, PixA_obj=None, Hmask=None, grid_pixsize=None, \
        count_thresh=None, canny_sig=None, peak_clip=0.7, VERBOSE_LEVEL=2):
        
        """
        # MeLon Notes 
        # What is Hough Transform?
        # a) A stright line in original-space can be geometrically charaterized by a certain pair (theta, rho) in hough-space. 
        #    The object stright line could be represented by x*cos(theta) + y*sin(theta) = rho
        #    The projection from a straight line in original-space to a point in hough-space is a bijection.
        #    
        #    [-] theta [-90d, +90d) is the rotate angle from X+ to the connecting line.
        #        > connecting line: line crossing the origin and the closest point on the straight line.
        #        > there are two directions of the connection line to measure rotate angle, we always 
        #          use the one yielding smaller angle.
        #        > the sign of theta: rotation towards Y+ (Y-) gets postive (negative) angles. 
        #          special case: if the connection line is orthogonal to X+, we adopt rotate angle of -90d.
        #
        #    [-] rho (-inf, +inf) is the length of connecting arrow (however, negative values allowed).
        #        > connecting arrow: arrow from the origin to the closest point on the straight line.
        #        > the sign of rho: arrow towards X+ (X-) get positive (negative) lengths.
        #          special case (theta = -90d): arrow towards Y- (Y+) get positive (negative) lengths.
        #
        #    [*] we can get the distance between two parallel straight lines using hough representations
        #        says, the hough representations are given as below, then their distance is |rho2 - rho1| geometrically.
        #        (1) x*cos(theta) + y*sin(theta) = rho1 | (2) x*cos(theta) + y*sin(theta) = rho2
        #        the corollary is that the distance of any point (x_t, y_t) to a given straight line (1) is
        #        |x_t*cos(theta) + y_t*sin(theta) - rho1|
        #
        #    Remarks on symmetric definition
        #    > There is a symmetric definition of (theta, rho), for which the equation becomes x*sin(theta) + y*cos(theta) = rho.
        #      For paragraphs [-]x2, only we need to change is replacing X+/X-/Y+/Y- by Y+/Y-/X+/X-.
        #      For paragraph [*], just swap the functions cos and sin.
        #      NOTE: Since scikit-image adopts this convention, we will stick with this kind of definition in our context!
        #
        #    Remarks on disadvantage of hough transform
        #    > one obvious disadvantage of hough transform is we don't know the start-end point of the detected stright line.
        #      probability hough transform can address this issue but more time-consuming.
        #
        # How does Hough Transform work?
        # b1) Consider a point A in orinal-space, all possible stright lines cross the point A will produce 
        #     a continuous curve U in hough-space, accordingly. 
        #  
        # b2) Assume two points A and B in orinal-space that can produce curves H(A) and H(B) in hough-space, respectively.
        #     Let K be the crosspoint of H(A) and H(B), it's easy to verify that K is the hough projection of 
        #     the stright line L determied by points A and B.
        #
        # b3) Imagine an additional point C appears on stright line L, its hough curve has to pass through K.
        #     As a result, more points on stright line L will accumulatively contribute to the crosspoint K. 
        #     P.S. Let K' be the crosspoint of H(A) and H(C), then K' is also the hough projection of the stright line L' 
        #          determined by A and C. Since L' = L then K' = K.
        #
        # c) Says, we have a set of scatter points in original-space, each point can produce a curve in hough-space.
        #    A 'high-density' region in hough-space would indicate that a corrsponding striaight feature exist in original-space.
        #
        # d) In practice, both original-space and hough-space are pixelized. One can detect stright line features 
        #    in original-space by the peaks in hough space. using count-threshold or canny edge detection are 
        #    popular preliminary processing to create binary image for hough detection.
        #    
        #    Note that the coordinate origin of a pixeled image is 
        #    (1) lying view: the lower-left corner of the lower-left (first) pixel. 
        #    (2) standing view: the top-left corner of the top-left (first) pixel. 
        # 
        # More details about scikit-image hough transform
        # e) scikit-image updates their implementation of hough transform since v0.19.0
        #    I tested hough transform to detect a thin stripe with random width 1.0-4.0 pix on 1024 X 1024 image.
        #    each row of the following listed as: input theta, input rho, detected-input rho, detected-input theta
        #
        #    > old versions 0.16.1 - 0.18.3: 
        #      [87.5d, 89.5d), [5, 1000], 0.28 +/- 2.87, 0.00 +/- 0.38        | near-vertical-1
        #      [-89.5d, -87.5d], [5, 1000], -1.25 +/- 5.03, -0.07 +/- 0.42    | near-vertical-2
        #      [-89.5d, -87.5d], [-1000, -5], 0.83 +/- 3.00, -0.00 +/- 0.36   | near-vertical-3
        #
        #      [45d, 55d], [5, 1000], 0.04 +/- 0.55, -0.00 +/- 0.36           | normal-1
        #      [-55d, -45d], [5, 1000], 1.32 +/- 5.18, 0.06 +/- 0.40          | normal-2
        #      [-55d, -45d], [-1000, -5], 0.93 +/- 4.64, 0.03 +/- 0.38        | normal-3
        #
        #      [-2.5d, -0.5d], [5, 1000], 0.19 +/- 3.08, -0.00 +/- 0.37       | near-horizental-1
        #      [-0.5d, 0.5d], [5, 1000], 0.54 +/- 2.99, 0.03 +/- 0.38         | near-horizental-2
        #      [0.5d, 2.5d], [5, 1000], -0.07 +/- 2.90, -0.03 +/- 0.37        | near-horizental-3
        #
        #    > new versions 0.19.0 - 0.19.3 (latest)
        #      [87.5d, 89.5d), [5, 1000], -0.51 +/- 2.85, -0.00 +/- 0.37      | near-vertical-1
        #      [-89.5d, -87.5d], [5, 1000], -0.16 +/- 4.98, 0.03 +/- 0.36     | near-vertical-2
        #      [-89.5d, -87.5d], [-1000, -5], 0.31 +/- 2.87, -0.02 +/- 0.34   | near-vertical-3
        #
        #      [45d, 55d], [5, 1000], -0.64 +/- 0.58, 0.00 +/- 0.37           | normal-1
        #      [-55d, -45d], [5, 1000], 0.48 +/- 4.87, 0.05 +/- 0.41          | normal-2
        #      [-55d, -45d], [-1000, -5], 0.18 +/- 4.78, 0.01 +/- 0.40        | normal-3
        #
        #      [-2.5d, -0.5d], [5, 1000], -0.69 +/- 2.98, -0.02 +/- 0.36      | near-horizental-1
        #      [-0.5d, 0.5d], [5, 1000], -0.55 +/- 3.02, -0.00 +/- 0.38       | near-horizental-2  
        #      [0.5d, 2.5d], [5, 1000], -0.54 +/- 2.84, -0.00 +/- 0.37        | near-horizental-3 
        #
        """
        
        # * check version
        skimage_version = parse_version(skimage.__version__)
        if skimage_version < parse_version('0.16.1'):
            if VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = 'current scikit-image [%s] is too old and not tested!' %skimage_version
                warnings.warn('MeLOn WARNING: %s' %_warn_message)
        
        if parse_version('0.16.1') <= skimage_version <= parse_version('0.18.3'):
            if VERBOSE_LEVEL in [2]:
                _message = 'current scikit-image [%s] uses classic implementation of hough transform' %skimage_version
                print('MeLOn CheckPoint: %s' %_message)
        
        if skimage_version >= parse_version('0.19.0'):
            if VERBOSE_LEVEL in [2]:
                _message = 'current scikit-image [%s] uses updated implementation of hough transform' %skimage_version
                print('MeLOn CheckPoint: %s' %_message)
        
        # * convert scatter points to be 2D-image
        if XY_obj is not None:
            """
            # Remarks on the pixelization
            # [-] the coordinate origin of the image is at (x_min, y_min), and the image can fully cover all scatter points.
            #     we therefore define the complete nodes (containing all boundaries) of the image grid as 
            #     arange(x_min, x_max + 2Gx, Gx), where Gx is grid pixel size along X axis
            #     arange(y_min, y_max + 2Gy, Gy), where Gx is grid pixel size along Y axis
            #
            # [-] when the scatter points contribute the pixels of the image grid, we will regard the left boundary 
            #     and lower boundary of a pixel as solid lines but leave others dashed lines (lying view).
            #  
            # NOTE: to avoid too large pixeled image, it is useful to reject the outliers of input scatter points.
            #       this function allows users to provide a mask for the rejection.
            #
            """
            XY_Hobj = XY_obj.copy()
            if Hmask is not None: XY_Hobj = XY_obj[Hmask]
            x_min, x_max = np.min(XY_Hobj[:, 0]), np.max(XY_Hobj[:, 0])
            xnodes = np.arange(x_min, x_max+2*grid_pixsize, grid_pixsize)
            y_min, y_max = np.min(XY_Hobj[:, 1]), np.max(XY_Hobj[:, 1])
            ynodes = np.arange(y_min, y_max+2*grid_pixsize, grid_pixsize)
            PixA_inp = np.zeros((len(xnodes)-1, len(ynodes)-1))
            for x, y in XY_Hobj:
                r = bisect.bisect_right(xnodes, x)-1
                c = bisect.bisect_right(ynodes, y)-1
                PixA_inp[r, c] += 1
        else:
            assert PixA_obj is not None
            PixA_inp = PixA_obj
        
        # * create mask from 2D-image by count-threshold or canny egde detection
        assert (count_thresh is not None) or (canny_sig is not None)
        if count_thresh is not None: Mask_inp = PixA_inp >= count_thresh
        if canny_sig is not None: Mask_inp = feature.canny(PixA_inp, sigma=canny_sig)
        
        # * perform hough transform and searching peak in hough space
        Hspace, Theta, Rho = hough_line(Mask_inp.astype(int))
        ThetaPeaks, RhoPeaks = hough_line_peaks(Hspace, Theta, Rho, threshold=peak_clip*np.max(Hspace))[1:]
        
        # * switch to the raw coordinate origin and length unit
        # * calculate the scalar distance from scatter points to the detected lines

        ScaLineDIST = None
        if XY_obj is not None:
            ScaLineDIST = []
            for i in range(len(RhoPeaks)):
                RhoPeaks[i] = grid_pixsize*RhoPeaks[i] + x_min*np.sin(ThetaPeaks[i]) + y_min*np.cos(ThetaPeaks[i])
                dist = np.abs(np.sin(ThetaPeaks[i])*XY_obj[:, 0] + np.cos(ThetaPeaks[i])*XY_obj[:, 1] - RhoPeaks[i])
                ScaLineDIST.append(dist)
            ScaLineDIST = np.array(ScaLineDIST).T

        return PixA_inp, Hspace, ThetaPeaks, RhoPeaks, ScaLineDIST



"""
Remarks on Internal Packages Imports:
    from sfft.AutoSparsePrep import Auto_SparsePrep
    from sfft.AutoCrowdedPrep import Auto_CrowdedPrep
    from sfft.utils.SymmetricMatch import Symmetric_Match
    from sfft.utils.HoughMorphClassifier import Hough_MorphClassifier
    from sfft.utils.WeightedQuantile import TopFlatten_Weighted_Quantile
    from sfft.utils.meta.MultiProc import Multi_Proc
    from sfft.utils.meta.TimeoutKit import TimeoutAfter
    from sfft.utils.pyAstroMatic.PYSEx import PY_SEx
    from sfft.sfftcore.SFFTConfigure import SFFTConfigure
    from sfft.sfftcore.SFFTSubtract import GeneralSFFTSubtract
"""

__version__ = "1.6.4"
from .PureCupyCustomizedPacket import PureCupy_Customized_Packet
from .CustomizedPacket import Customized_Packet
from .AutoSparsePrep import Auto_SparsePrep
from .AutoCrowdedPrep import Auto_CrowdedPrep
from .EasySparsePacket import Easy_SparsePacket
from .EasyCrowdedPacket import Easy_CrowdedPacket
from .MultiEasySparsePacket import MultiEasy_SparsePacket
from .MultiEasyCrowdedPacket import MultiEasy_CrowdedPacket
from .BSplineSFFT import BSpline_Packet, BSpline_MatchingKernel, BSpline_DeCorrelation, BSpline_GridConvolve



import sep
import fastremap
import numpy as np
from astropy.io import fits
import scipy.ndimage as ndimage
from sfft.utils.pyAstroMatic.PYSEx import PY_SEx
from sfft.utils.WeightedQuantile import TopFlatten_Weighted_Quantile
# version: Feb 4, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Auto_CrowdedPrep:
    def __init__(self, FITS_REF, FITS_SCI, GAIN_KEY='GAIN', SATUR_KEY='SATURATE', BACK_TYPE='AUTO', BACK_VALUE=0.0, \
        BACK_SIZE=64, BACK_FILTERSIZE=3, DETECT_THRESH=5.0, ANALYSIS_THRESH=5.0, DETECT_MINAREA=5, DETECT_MAXAREA=0, \
        DEBLEND_MINCONT=0.005, BACKPHOTO_TYPE='LOCAL', ONLY_FLAGS=None, BoundarySIZE=0.0, VERBOSE_LEVEL=2):

        self.FITS_REF = FITS_REF
        self.FITS_SCI = FITS_SCI

        self.GAIN_KEY = GAIN_KEY
        self.SATUR_KEY = SATUR_KEY

        self.BACK_TYPE = BACK_TYPE
        self.BACK_VALUE = BACK_VALUE
        self.BACK_SIZE = BACK_SIZE
        self.BACK_FILTERSIZE = BACK_FILTERSIZE

        self.DETECT_THRESH = DETECT_THRESH
        self.ANALYSIS_THRESH = ANALYSIS_THRESH
        self.DETECT_MINAREA = DETECT_MINAREA
        self.DETECT_MAXAREA = DETECT_MAXAREA
        self.DEBLEND_MINCONT = DEBLEND_MINCONT
        self.BACKPHOTO_TYPE = BACKPHOTO_TYPE

        self.ONLY_FLAGS = ONLY_FLAGS
        self.BoundarySIZE = BoundarySIZE
        self.VERBOSE_LEVEL = VERBOSE_LEVEL
    
    def AutoMask(self, BACK_SIZE_SUPER=128, StarExt_iter=2, PriorBanMask=None):

        # * Generate Super-Background for REF & SCI
        PixA_REF = fits.getdata(self.FITS_REF, ext=0).T
        PixA_SCI = fits.getdata(self.FITS_SCI, ext=0).T

        if not PixA_REF.flags['C_CONTIGUOUS']:
            PixA_REF = np.ascontiguousarray(PixA_REF, np.float64)
        else: PixA_REF = PixA_REF.astype(np.float64)

        if not PixA_SCI.flags['C_CONTIGUOUS']:
            PixA_SCI = np.ascontiguousarray(PixA_SCI, np.float64)
        else: PixA_SCI = PixA_SCI.astype(np.float64)

        # NOTE: sep requires input_array.flags['C_CONTIGUOUS'] is True & input_array.dtype.byteorder is '='
        PixA_SBG_REF = sep.Background(PixA_REF, bw=BACK_SIZE_SUPER, bh=BACK_SIZE_SUPER, fw=3, fh=3).back()        
        PixA_SBG_SCI = sep.Background(PixA_SCI, bw=BACK_SIZE_SUPER, bh=BACK_SIZE_SUPER, fw=3, fh=3).back()

        # * Generate Saturation-Mask for REF & SCI
        def GenSatMask(FITS_obj):

            # ** trigger a very-cold SExtractor
            SExParam = ['X_IMAGE', 'Y_IMAGE', 'FLUX_AUTO', 'FLUXERR_AUTO', 'FLUX_MAX', 'FWHM_IMAGE']
            PYSEX_OP = PY_SEx.PS(FITS_obj=FITS_obj, SExParam=SExParam, GAIN_KEY=self.GAIN_KEY, \
                SATUR_KEY=self.SATUR_KEY, BACK_TYPE=self.BACK_TYPE, BACK_VALUE=self.BACK_VALUE, \
                BACK_SIZE=self.BACK_SIZE, BACK_FILTERSIZE=self.BACK_FILTERSIZE, DETECT_THRESH=self.DETECT_THRESH, \
                ANALYSIS_THRESH=self.ANALYSIS_THRESH, DETECT_MINAREA=self.DETECT_MINAREA, \
                DETECT_MAXAREA=self.DETECT_MAXAREA, DEBLEND_MINCONT=self.DEBLEND_MINCONT, \
                BACKPHOTO_TYPE=self.BACKPHOTO_TYPE, CHECKIMAGE_TYPE='SEGMENTATION', AddRD=False, \
                ONLY_FLAGS=self.ONLY_FLAGS, XBoundary=self.BoundarySIZE, YBoundary=self.BoundarySIZE, \
                MDIR=None, VERBOSE_LEVEL=self.VERBOSE_LEVEL)
            AstSEx, PixA_SEG = PYSEX_OP[0], PYSEX_OP[1][0].astype(int)
            
            # *** Note on the crude estimate of FWHM ***
            # our strategy is to calculate a median FWHM_IMAGE weighted by the source flux, meanwhile, 
            # we modulate the weights by applying a penalty on the sources with large FWHM_IMAGE.

            NTE_4FWHM = 30  # NUM_TOP_END for FWHM, default value used herein.
            _values, _weights = np.array(AstSEx['FWHM_IMAGE']), np.array(AstSEx['FLUX_AUTO'])
            _weights /= np.clip(_values, a_min=1.0, a_max=None)**2  
            FWHM = TopFlatten_Weighted_Quantile.TFWQ(values=_values, weights=_weights, \
                quantiles=[0.5], NUM_TOP_END=NTE_4FWHM)[0]

            # ** determine initial saturation region
            SATLEVEL = fits.getheader(FITS_obj, ext=0)[self.SATUR_KEY]
            AstSEx_SAT = AstSEx[AstSEx['FLUX_MAX'] >= SATLEVEL]
            Mappings = {}
            for label in AstSEx_SAT['SEGLABEL']: 
                Mappings[label] = -label
            fastremap.remap(PixA_SEG, Mappings, preserve_missing_labels=True, in_place=True)
            SatMask = PixA_SEG < 0
            
            # ** refine the saturation regions
            #    NOTE: SExtractor tend to label some outskirt-islands as part of a saturated source.

            XY_SAT = np.array([AstSEx_SAT['X_IMAGE'], AstSEx_SAT['Y_IMAGE']]).T            
            Lmap = ndimage.measurements.label(SatMask)[0]
            SATL = Lmap[((XY_SAT[:, 0] - 0.5).astype(int), (XY_SAT[:, 1] - 0.5).astype(int))]
            SATL = list(set(SATL).difference({0}))          # 0 is background
            SatMask = np.in1d(Lmap.flatten(), SATL).reshape(Lmap.shape)    # WARNING NOTE UPDATE

            struct0 = ndimage.generate_binary_structure(2, 1)
            struct = ndimage.iterate_structure(struct0, StarExt_iter)
            SatMask = ndimage.grey_dilation(SatMask, footprint=struct)     # WARNING NOTE UPDATE
            
            NUM_SAT = len(AstSEx_SAT)
            SatPROP = np.sum(SatMask) / (SatMask.shape[0] * SatMask.shape[1])
            return SATLEVEL, FWHM, SatMask, NUM_SAT, SatPROP

        SATLEVEL_REF, FWHM_REF, SatMask_REF, NUM_SAT_REF, SatPROP_REF = GenSatMask(self.FITS_REF)
        SATLEVEL_SCI, FWHM_SCI, SatMask_SCI, NUM_SAT_SCI, SatPROP_SCI = GenSatMask(self.FITS_SCI)

        if self.VERBOSE_LEVEL in [1, 2]:
            _message = 'Estimated [FWHM_REF = %.3f pix] & [FWHM_SCI = %.3f pix]!' %(FWHM_REF, FWHM_SCI)
            print('\nMeLOn CheckPoint: %s' %_message)

        if self.VERBOSE_LEVEL in [2]:
            _message = 'The SATURATED Regions --- Number (Pixel Proportion) '
            _message += '[REF = %d (%s)] & ' %(NUM_SAT_REF, '{:.2%}'.format(SatPROP_REF))
            _message += '[SCI = %d (%s)]!' %(NUM_SAT_SCI, '{:.2%}'.format(SatPROP_SCI))
            print('\nMeLOn CheckPoint: %s' %_message)
        
        # * Define ProhibitedZone
        NaNmask_U = None
        if PriorBanMask is None:
            ProZone = np.logical_or(SatMask_REF, SatMask_SCI)
        else: ProZone = np.logical_or.reduce((PriorBanMask, SatMask_REF, SatMask_SCI))
        
        NaNmask_REF = np.isnan(PixA_REF)
        NaNmask_SCI = np.isnan(PixA_SCI)
        if NaNmask_REF.any() or NaNmask_SCI.any():
            NaNmask_U = np.logical_or(NaNmask_REF, NaNmask_SCI)
            ProZone[NaNmask_U] = True

        # * Make ActiveMask > PixA_mREF & PixA_mSCI
        PixA_mREF = PixA_REF.copy()
        PixA_mSCI = PixA_SCI.copy()
        
        # NOTE: The preparation can guarantee that mREF & mSCI are NaN-Free !
        PixA_mREF[ProZone] = PixA_SBG_REF[ProZone]      # NOTE REF is not sky-subtracted
        PixA_mSCI[ProZone] = PixA_SBG_SCI[ProZone]      # NOTE SCI is not sky-subtracted

        ActiveMask = ~ProZone
        if self.VERBOSE_LEVEL in [1, 2]:
            ActivePROP = np.sum(ActiveMask) / (ActiveMask.shape[0] * ActiveMask.shape[1])
            print('\nMeLOn CheckPoint: Active-Mask Pixel Proportion [%s]' %('{:.2%}'.format(ActivePROP)))

        # * Create sfft-prep master dictionary
        SFFTPrepDict = {}
        SFFTPrepDict['PixA_REF'] = PixA_REF
        SFFTPrepDict['PixA_SCI'] = PixA_SCI
        SFFTPrepDict['Union-NaN-Mask'] = NaNmask_U

        SFFTPrepDict['SATLEVEL_REF'] = SATLEVEL_REF
        SFFTPrepDict['SATLEVEL_SCI'] = SATLEVEL_SCI
        SFFTPrepDict['FWHM_REF'] = FWHM_REF
        SFFTPrepDict['FWHM_SCI'] = FWHM_SCI

        SFFTPrepDict['REF-SAT-Mask'] = SatMask_REF
        SFFTPrepDict['SCI-SAT-Mask'] = SatMask_SCI
        SFFTPrepDict['Active-Mask'] = ActiveMask
        SFFTPrepDict['PixA_mREF'] = PixA_mREF
        SFFTPrepDict['PixA_mSCI'] = PixA_mSCI

        return SFFTPrepDict



import numpy as np
from astropy import units as u
from scipy.spatial import cKDTree
from astropy.coordinates import SkyCoord
from astropy.coordinates import match_coordinates_sky
# version: Aug 4, 2022

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.2"

class Symmetric_Match:
    @staticmethod
    def SM(XY_A, XY_B, tol, return_distance=False):

        """
        a) XY_A, XY_B are two collections of points for matching with shape (NUM_A, 2) and (NUM_B, 2)
        b) The function func_query(PO, POR) returns the matched points for PO in POR.
           For any given po in PO, assign the index of the nearest point in POR (respect to po), 
           and if the nearest distance still exceeds the tolerance tol, assign the trivial index POR.shape[0].
        c) A point pair < a, b > is called symmetrically matched iff they are nearest to each other.
        """

        def func_query(PO, POR):
            # Find k-nearest (here k=1) objects in POR
            # [0] distance to the POR object
            # [1] index of the POR object, if index = POR.shape[0] 
            #     the query fails within given tolerance.
            Tree = cKDTree(POR)
            DIST, IDX = Tree.query(PO, k=1, distance_upper_bound=tol)
            return DIST, IDX
        
        NUM_A, NUM_B = XY_A.shape[0], XY_B.shape[0]
        DIST_alpha, IDX_alpha = func_query(PO=XY_A, POR=XY_B)
        DIST_beta, IDX_beta = func_query(PO=XY_B, POR=XY_A)
        
        MATCH_IDX_A, MATCH_IDX_B, MATCH_DIST = [], [], []
        if NUM_A < NUM_B:
            for i in np.where((IDX_alpha < NUM_B))[0]:
                j = IDX_alpha[i]
                if IDX_beta[j] == i:
                    MATCH_IDX_A.append(i)
                    MATCH_IDX_B.append(j)
                    MATCH_DIST.append(DIST_alpha[i])
        else:
            for v in np.where((IDX_beta < NUM_A))[0]:
                u = IDX_beta[v]
                if IDX_alpha[u] == v:
                    MATCH_IDX_A.append(u)
                    MATCH_IDX_B.append(v)
                    MATCH_DIST.append(DIST_beta[v])
        
        Symm = np.array([MATCH_IDX_A, MATCH_IDX_B]).T
        MATCH_DIST = np.array(MATCH_DIST)
        
        if return_distance:
            return Symm, MATCH_DIST
        else:
            return Symm

class Sky_Symmetric_Match:
    @staticmethod
    def SSM(RD_A, RD_B, tol, return_distance=False):
        
        # NOTE: SkyCoord default frame is icrs, only having subtle difference with (equinox='J2000', frame='fk5')
        Skycoors_A = SkyCoord(ra=RD_A[:, 0], dec=RD_A[:, 1], unit=(u.deg, u.deg))
        Skycoors_B = SkyCoord(ra=RD_B[:, 0], dec=RD_B[:, 1], unit=(u.deg, u.deg))
        
        NUM_A, NUM_B = RD_A.shape[0], RD_B.shape[0]
        if NUM_A < NUM_B:
            _matchres = match_coordinates_sky(Skycoors_A.frame, Skycoors_B.frame)[:2]
            IDX_alpha, DIST_alpha = _matchres[0], _matchres[1].arcsec
            AIDX_cone = np.where(DIST_alpha < tol)[0]
            BIDX_conep = IDX_alpha[AIDX_cone]
            DIST_conep = DIST_alpha[AIDX_cone]
            
            IDX_beta = match_coordinates_sky(Skycoors_B[BIDX_conep].frame, Skycoors_A.frame)[0]
            SYMM_MASK = AIDX_cone == IDX_beta
            MATCH_IDX_A = AIDX_cone[SYMM_MASK]
            MATCH_IDX_B = BIDX_conep[SYMM_MASK]
            MATCH_DIST = DIST_conep[SYMM_MASK]
        else:
            _matchres = match_coordinates_sky(Skycoors_B.frame, Skycoors_A.frame)[:2]
            IDX_beta, DIST_beta = _matchres[0], _matchres[1].arcsec
            BIDX_cone = np.where(DIST_beta < tol)[0]
            AIDX_conep = IDX_beta[BIDX_cone]
            DIST_conep = DIST_beta[BIDX_cone]
            
            IDX_alpha = match_coordinates_sky(Skycoors_A[AIDX_conep].frame, Skycoors_B.frame)[0]
            SYMM_MASK = BIDX_cone == IDX_alpha
            MATCH_IDX_B = BIDX_cone[SYMM_MASK]
            MATCH_IDX_A = AIDX_conep[SYMM_MASK]
            MATCH_DIST = DIST_conep[SYMM_MASK]
        
        Symm = np.array([MATCH_IDX_A, MATCH_IDX_B]).T

        if return_distance:
            return Symm, MATCH_DIST
        else:
            return Symm



import re
import os
from tempfile import mkdtemp
# version: Jan 21, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class AMConfig_Maker:
    @staticmethod
    def AMCM(MDIR, AstroMatic_KEY, ConfigDict=None, tag='pyastromatic'):

        assert AstroMatic_KEY in ['sex', 'sextractor', 'psfex', 'swarp', 'scamp', 'sky']
        prefix, suffix = '%s_' %tag, '%s_config' %AstroMatic_KEY
        TDIR = mkdtemp(suffix=suffix, prefix=prefix, dir=MDIR)
        config_path = ''.join([TDIR, "/%s.%s" %(tag, AstroMatic_KEY)])

        """
        # Remarks on missfits (not included here)
        # [1] Use missfits to split Multi-Extension-FITS can be simply done in terminal
        #     e.g., missfits -OUTFILE_TYPE SPLIT -WRITE_XML N FITS_MEF
        #
        # [2] Use missfits remove keyword in header
        #     e.g., missfits X.fits -REMOVE_KEYWORD SKY_BACK -SAVE_TYPE REPLACE        
        #     e.g., missfits X.fits -REMOVE_KEYWORD SKY_BACK -SAVE_TYPE NEW -NEW_SUFFIX .miss
        #
        """

        # * Creat default configuration file and extract keywords
        os.system("%s -dd > %s" %(AstroMatic_KEY, config_path))    # WARNING: may be invalid for other os-platform
        config_lst = open(config_path).read().split('\n')
        Keys = [re.split("\s+", line)[0] for line in config_lst]   # WARNING: invalid lines will get '' or '#'

        # * Modify the configuration 
        for key in ConfigDict:
            value = ConfigDict[key]
            if isinstance(value, tuple):
                value = ','.join([str(v) for v in value])
            else: value = str(value)
            
            index = Keys.index(key)
            line = config_lst[index]
            
            comment = ''
            cindx = re.search('#', line)
            if cindx is not None:
                comment = line[cindx.start():]
            config_lst[index] = '    '.join([key, value, comment])

        string = "\n".join(config_lst)
        sfile = open(config_path, 'w')
        sfile.write(string)
        sfile.close()

        return config_path



import skimage
import warnings
import numpy as np
from pkg_resources import parse_version
from sfft.utils.pyAstroMatic.PYSEx import PY_SEx
from sfft.utils.HoughDetection import Hough_Detection
from sfft.utils.WeightedQuantile import TopFlatten_Weighted_Quantile

# version: Feb 10, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Hough_MorphClassifier:
    
    def __init__(self):
        
        """
        # MeLOn Notes
        # @ Point-Source Extractor
        #   A) A PSFEx suggested Morphological Classifier, based on a 2D distribution diagram
        #      FLUX_RADIUS [X-axis] - MAG_AUTO [Y-axis], A universal but naive approach.
        #      We first draw all isolated sources on the plane, and the typical distribution will form a 'Y' shape.
        #      A nearly vertical branch, A nearly horizontal branch and their cross with a tail at faint side.
        #
        #      Here I give rough conclusions with comments, note I have compared with Legacy Survety Tractor Catalog.
        #      a. The point sources would be distributed around the nearly vertical stright line. {vertical branch}
        #         NOTE At the faint end, point sources no longer cling to the line, being diffuse in the cross and tail.
        #      b. Close to the bright end of the stright-line are saturated or slight-nonlinear sources, with a deviated direction.
        #      c. The right side of the line are typically extended structure, mainly including various galaxies. {horizontal branch}
        #         NOTE At the faint end, likewise, extended sources also exist, being diffuse in the cross and tail.
        #      d. Scattering points are located at very left side of the line, they are generally hotpix, cosmic-ray or 
        #         some small-scale artifacts. Keep in mind, they are typically outlier-like and away from the cross and tail.
        #
        #      METHOD: For simplicity, we only crudely divide the diagram into 3 regions, w.r.t. the vertical line.
        #              they are, Radius-Mid (FR-M), Radius-Large (FR-L) and Radius-Small (FR-S).
        #
        #   B) 3 hierarchic groups                
        #      > "Good" Sources:
        #        + NOT in FR-S region (union of FR-M & FR-L)
        #          NOTE Good Sources is consist of the vertical & horizontal branches with their cross (not the tail), 
        #               which is roughly equivalent to the set of REAL Point-Sources & Extended Sources 
        #               with rejection the samples in the tail (at faint & small-radius end).
        #          NOTE Good Sources are commonly used as FITTING Candidates in Image Subtraction.
        #               It is acceptable to lose the samples in the tail.
        #         
        #        >> {subgroup} "Point-like" Sources:
        #               + Restricted into FR-M Region   |||   Should be located around the Hough-Line
        #               + Basically Circular-Shape      |||   PSFEx-ELLIPTICITY = (A-B) / (A+B) < PointSource_MINELLIP
        #
        #                 NOTE At cross region, this identification criteria mis-include some extended source
        #                      On the flip side, some REAL PointSource samples are missing in the tail.
        #                 NOTE Point Sources are usually employed as FWHM Estimator.
        #                 NOTE We may lossen PointSource_MINELLIP if psf itself is significantly asymmetric (e.g. tracking problem).
        #
        #                 WARNING: At the faint end, this subgroup would be contaminated by round extended sources 
        #                          (meet PointSource_MINELLIP) which also lie in the cross region. Especially when the field 
        #                          is galaxy-dominated, the contamination can be considerable. In light of the fact that 
        #                          cross region is located at the faint end in the belt, we calculate the median FWHM 
        #                          weighted by source flux, to suppress the over-estimate trend driven by the extended sources.
        #
        #   C) Additional WARNINGS
        #      a. This extracor is ONLY designed for sparse field (isolated sources dominated cases).
        #         We generally only take these isloated & non-saturated sources (FLAGS = 0) into account.
        #         But one may also loosen the constrain on FLAGS to be [0,2].
        #
        #      b. We employ Hough Transform to detect the Stright-Line feature in the image, 
        #         naturally sampled from the raw scatter diagram. But note such diagram makes sense 
        #         only if we could detect enough sources (typically > 200) in the given image.
        #         NOTE Reversed axes employed --- MAG_AUTO [X-axis] - FLUX_RADIUS [Y-axis].
        #
        """

        pass

    def MakeCatalog(FITS_obj, GAIN_KEY='GAIN', SATUR_KEY='SATURATE', BACK_TYPE='AUTO', BACK_VALUE=0.0, \
        BACK_SIZE=64, BACK_FILTERSIZE=3, DETECT_THRESH=1.5, ANALYSIS_THRESH=1.5, DETECT_MINAREA=5, \
        DETECT_MAXAREA=0, DEBLEND_MINCONT=0.005, BACKPHOTO_TYPE='LOCAL', CHECKIMAGE_TYPE='NONE', \
        AddRD=False, ONLY_FLAGS=[0], BoundarySIZE=30, AddSNR=True, VERBOSE_LEVEL=2):

        """
        # * Remarks on the SExtractor Catalog for Hough Transform
        #   [1] Although XYWIN can be more accurate than XY for point sources, here 
        #       we adopt XY rather than XYWIN for both point and extended sources as a trade-off.
        #   [2] by default, it only takes isolated & non-saturated sources (FLAGS = 0) into account.
        #   [3] one may need to tune DETECT_THRESH & DETECT_MINAREA for a specific program.
        #
        """

        SExParam = ['X_IMAGE', 'Y_IMAGE', 'FLUX_AUTO', 'FLUXERR_AUTO', 'MAG_AUTO', 'MAGERR_AUTO', \
                     'FLAGS', 'FLUX_RADIUS', 'FWHM_IMAGE', 'A_IMAGE', 'B_IMAGE']
        
        if AddSNR: SExParam.append('SNR_WIN')
        PYSEX_OP = PY_SEx.PS(FITS_obj=FITS_obj, SExParam=SExParam, GAIN_KEY=GAIN_KEY, SATUR_KEY=SATUR_KEY, \
            BACK_TYPE=BACK_TYPE, BACK_VALUE=BACK_VALUE, BACK_SIZE=BACK_SIZE, BACK_FILTERSIZE=BACK_FILTERSIZE, \
            DETECT_THRESH=DETECT_THRESH, ANALYSIS_THRESH=ANALYSIS_THRESH, DETECT_MINAREA=DETECT_MINAREA, \
            DETECT_MAXAREA=DETECT_MAXAREA, DEBLEND_MINCONT=DEBLEND_MINCONT, BACKPHOTO_TYPE=BACKPHOTO_TYPE, \
            CHECKIMAGE_TYPE=CHECKIMAGE_TYPE, AddRD=AddRD, ONLY_FLAGS=ONLY_FLAGS, XBoundary=BoundarySIZE, \
            YBoundary=BoundarySIZE, MDIR=None, VERBOSE_LEVEL=VERBOSE_LEVEL)
        
        return PYSEX_OP
    
    def Classifier(AstSEx, Hough_MINFR=0.1, Hough_MAXFR=10.0, Hough_PeakClip=0.7, \
        BeltHW=0.2, PointSource_MINELLIP=0.3, VERBOSE_LEVEL=2):

        """
        # * calssifier based on hough detection  
        #   We use hough transform to detect the point source belt from the scatter points 
        #   (X: MAG_AUTO, Y: FLUX_RADIUS). The belt is a nearly horizontal straight line.
        #
        #   NOTE: the point source belt has a typical theta of 1.5d +/- 1.0d. on this occasion, 
        #         we should stick with the classic implementation of hough transform in 
        #         scikit-image [0.16.1 to 0.18.3], where the measurement bias on rho
        #         is closer to zero than the new implementation!
        #
        #   NOTE: in general, the point source belt is the strongest line feature. 
        #         by default, -Hough_PeakClip=0.7 is proper in most cases.
        #         however, for galaxy dominated fields (e.g., JWST imagines of galaxy cluster),
        #         the point-source-belt can be not very pronounced, and one may should reduce 
        #         the parameter from default 0.7 to, says, ~ 0.4.
        #
        #   Remarks on the Mask
        #   It can be essential to apply restrictions on FLUX_RADIUS (R) for the scatter points used in hough transform. 
        #   (1) Exclude the sources with unusally large R (e.g., R > 20.0) can speed up the calculations.
        #       One can control the upper limit of R using parameter -Hough_MAXFR
        #
        #   (2) The sources with very small R (e.g., R < 0.5) may become districtions for hough detection.
        #       They are not point sources, likely hot pixels or cosmic rays, sometimes,
        #       wrong line features will be detected by chance without proper rejections.
        #
        #       NOTE: One can control the lower limit of R via parameter -Hough_MINFR,
        #             the recommended values of Hough_MINFR range from 0.1 to 1.0. One should choose 
        #             a proper value according to the instrument configuration and typical seeing conditions.
        #
        """

        skimage_version = parse_version(skimage.__version__)
        assert parse_version('0.16.1') <= skimage_version <= parse_version('0.18.3')
        
        A_IMAGE = np.array(AstSEx['A_IMAGE'])
        B_IMAGE = np.array(AstSEx['B_IMAGE'])
        MA_FR = np.array([AstSEx['MAG_AUTO'], AstSEx['FLUX_RADIUS']]).T
        
        ELLIP = (A_IMAGE - B_IMAGE) / (A_IMAGE + B_IMAGE)
        if PointSource_MINELLIP is not None:
            MASK_ELLIP = ELLIP < PointSource_MINELLIP
        else: MASK_ELLIP = np.ones(len(ELLIP)).astype(bool)

        MA, FR = MA_FR[:, 0], MA_FR[:, 1]
        MA_MID = np.nanmedian(MA)
        
        magnitude_main_range = (-7.0, 7.0)   # emperical
        Hmask = np.logical_and.reduce((FR > Hough_MINFR, FR < Hough_MAXFR, \
            MA > MA_MID + magnitude_main_range[0], MA < MA_MID + magnitude_main_range[1]))

        grid_pixsize, count_thresh = 0.05, 1   # emperical
        _res = Hough_Detection.HD(XY_obj=MA_FR, Hmask=Hmask, grid_pixsize=grid_pixsize, \
            count_thresh=count_thresh, peak_clip=Hough_PeakClip)
        ThetaPeaks, RhoPeaks, ScaLineDIST = _res[2:]

        BeltTheta_thresh = 0.2   # emperical
        NHOR_INDEX = np.where(np.abs(ThetaPeaks) < BeltTheta_thresh)[0]
        
        if len(NHOR_INDEX) == 0: 
            bingo_index = None
            if VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = '[NO] near-horizon peak as Point-Source-Belt!'
                warnings.warn('MeLOn WARNING: %s' %_warn_message)

        elif len(NHOR_INDEX) == 1:
            bingo_index = NHOR_INDEX[0]
            if VERBOSE_LEVEL in [1, 2]:
                _message = '[UNIQUE] near-horizon peak as Point-Source-Belt!'
                print('MeLOn CheckPoint: %s' %_message)
        
        else:
            bingo_index = np.min(NHOR_INDEX)
            if VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = '[MULTIPLE] near-horizon peaks, of which [strongest] as Point-Source-Belt!'
                warnings.warn('MeLOn WARNING: %s' %_warn_message)
        
        if bingo_index is not None:
            # * use hough detection method
            BeltTheta = ThetaPeaks[bingo_index]
            BeltRho = RhoPeaks[bingo_index]
            MASK_FRM = ScaLineDIST[:, bingo_index] < BeltHW

            if VERBOSE_LEVEL in [2]:
                _message = 'The Point-Source-Belt detected by Hough Transform '
                _message += 'is characterized by [%f, %f]!' %(BeltTheta, BeltRho)
                print('MeLOn CheckPoint: %s' %_message)

            # *** Note on the FRL region *** 
            # geometrically, we know the scatter points above the near-horizon straight line have
            # x*sin(theta) + y*cos(theta) > rho, where theta is BeltTheta and rho is BeltRho
            
            MASK_FRL = MA_FR[:, 0]*np.sin(BeltTheta) + MA_FR[:, 1]*np.cos(BeltTheta) > BeltRho
            MASK_FRL = np.logical_and(MASK_FRL, ~MASK_FRM)  # exlcude the points in the FR-M region.

        else:
            # * use standby method
            BeltTheta, BeltRho = np.nan, np.nan
            if VERBOSE_LEVEL in [0, 1, 2]:
                _warn_message = '[STANDBY] method to determine FR-S/M/L regions!'
                warnings.warn('MeLOn WARNING: %s' %_warn_message)

            # *** Note on the standby belt determination *** 
            # it is quite tricky to find a generic way to determine the point source belt when 
            # hough detection does not work. we noticed that the bright sources with moderate 
            # Flux Radius are more likely point sources. a standby strategy is to calculate 
            # a median Flux Radius weighted by the source flux, and meanwhile, we modulate 
            # the weights by applying a penalty on the sources with large Flux Radius.

            NTE_4FRM = 30  # NUM_TOP_END for FRM, default value used herein.
            _values, _weights = MA_FR[:, 1], np.array(AstSEx['FLUX_AUTO'])
            _weights /= np.clip(_values, a_min=0.5, a_max=None)**2  
            
            FR_MID = TopFlatten_Weighted_Quantile.TFWQ(values=_values, \
                weights=_weights, quantiles=[0.5], NUM_TOP_END=NTE_4FRM)[0]
            
            MASK_FRM = np.abs(MA_FR[:, 1]  - FR_MID) < BeltHW
            MASK_FRL = MA_FR[:, 1]  - FR_MID > BeltHW

        MASK_FRS = ~np.logical_or(MASK_FRM, MASK_FRL)
        LABEL_FR = np.array(['FR-S'] * len(AstSEx))
        LABEL_FR[MASK_FRM] = 'FR-M'
        LABEL_FR[MASK_FRL] = 'FR-L'

        if VERBOSE_LEVEL in [2]:
            _message = 'HoughMorphClassifier FR-S/M/L counts are '
            _message += '[%d / %d / %d]' %(np.sum(MASK_FRS), np.sum(MASK_FRM), np.sum(MASK_FRL))
            print('MeLOn CheckPoint: %s' %_message)

        # * determine Good Sources
        MASK_GS = ~MASK_FRS

        # * determine Point Sources
        MASK_PS = np.logical_and(MASK_FRM, MASK_ELLIP)
        if VERBOSE_LEVEL in [1, 2]:
            print('MeLOn CheckPoint: [%d] Good-Sources on the Image!' %np.sum(MASK_GS))
            print('MeLOn CheckPoint: [%d] Point-Sources on the Image!' %np.sum(MASK_PS))

        # * estimate FWHM
        NTE_4FWHM = 30  # NUM_TOP_END for FWHM, default value used herein.
        _values, _weights = np.array(AstSEx[MASK_PS]['FWHM_IMAGE']), np.array(AstSEx[MASK_PS]['FLUX_AUTO'])
        FWHM = TopFlatten_Weighted_Quantile.TFWQ(values=_values, weights=_weights, \
            quantiles=[0.5], NUM_TOP_END=NTE_4FWHM)[0]
        
        FWHM = round(FWHM, 6)
        if VERBOSE_LEVEL in [1, 2]:
            print('MeLOn CheckPoint: Estimated [FWHM = %.3f pix] from Point-Sources' %FWHM)

        return BeltTheta, BeltRho, LABEL_FR, MASK_GS, MASK_PS, FWHM
